# BSD 2-Clause License

# Copyright (c) 2018, Zachariah Etienne All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# indexedexpNRPyPN.py: functions related to indexed expressions,
# including e.g., tensors and pseudotensors.
# *** This is a stripped-down version of the indexedexp.py module
#     in the NRPy+ root directory.

# Step 1: Load needed modules
import sympy as sp  # SymPy: The Python computer algebra package upon which NRPy+ depends


def zerorank1(DIM=-1):
    if DIM == -1:
        DIM = 3  # default to 3D
    return [sp.sympify(0) for i in range(DIM)]


def zerorank3(DIM=-1):
    if DIM == -1:
        DIM = 3  # default to 3D
    return [
        [[sp.sympify(0) for i in range(DIM)] for j in range(DIM)]
        for k in range(DIM)
    ]


def declarerank1(objname, DIM=-1):
    if DIM == -1:
        DIM = 3  # default to 3D
    return [sp.sympify(objname + str(i)) for i in range(DIM)]


class NonInvertibleMatrixError(ZeroDivisionError):
    """Matrix Not Invertible; Division By Zero"""


# Define the rank-3 version of the Levi-Civita symbol.
def LeviCivitaSymbol_dim3_rank3():
    LeviCivitaSymbol = zerorank3(DIM=3)

    for i in range(3):
        for j in range(3):
            for k in range(3):
                # From https://codegolf.stackexchange.com/questions/160359/levi-civita-symbol :
                LeviCivitaSymbol[i][j][k] = (
                    (i - j) * (j - k) * (k - i) * sp.Rational(1, 2)
                )
    return LeviCivitaSymbol



# BSD 2-Clause License

# Copyright (c) 2018, Zachariah Etienne All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# As documented in the NRPyPN notebook
# PN-p_r.ipynb, this Python script
# generates the expression for the radial
# component of momentum p_r up to and
# including terms at 3.5PN order.
# It largely follows the technique of
#  Ramos-Buades, Husa, and Pratten (2018)
#    https://arxiv.org/abs/1810.00036

# Core functions:
# f_Htot_xyplane_binary(m1,m2, n12U,n21U, S1U, S2U, p1U,p2U, q)
#        Given standard input parameters, compute
#        the Hamiltonian for a binary system
#        orbiting instantaneously on the xy plane,
#        and store to the global variable
#        Htot_xyplane_binary

# f_dr_dt(Htot_xyplane_binary, m1,m2, n12U,n21U, chi1U,chi2U, S1U,S2U, p1U,p2U, r)
#         Given Htot_xyplane_binary (computed
#         above) and other standard input
#         parameters, compute
#         dr_dt = dr/dt and store to global
#         variable of the same name.

# f_p_r(m1,m2, chi1U,chi2U, r)
#       Compute p_r and store to
#       global variable of the same name.

# f_p_r_fullHam(m1,m2, n12U,n21U, chi1U,chi2U, S1U,S2U, p1U,p2U, r)
#       Compute p_r using the full
#       Hamiltonian, without truncating
#       higher-order terms self-consistently.

# Author:  Zach Etienne
#          zachetie **at** gmail **dot* com

# Step 0: Add NRPy's directory to the path
# https://stackoverflow.com/questions/16780014/import-file-from-parent-directory
import sympy as sp  # SymPy: The Python computer algebra package upon which NRPy+ depends

from .NRPyPN_shortcuts import (  # NRPyPN: shortcuts for e.g., vector operations
    Pr,
    Pt,
    div,
    nU,
)

#################################
#################################
# Basic equation for p_r can be written in the
#   form:
# p_r \approx [dr/dt - (partial_H/partial_{p_r})|_{p_r=0}] * [(partial^2_{H}/partial_{p_r^2})|_{p_r=0}]^{-1},
#  where
# dr/dt = [dE_{\rm GW}/dt + dM/dt] * [dH_{circ} / dr]^{-1},
#  and
# H_{circ} = Htot_xyplane_binary|_{p_r=0}
#  -> [dH_{circ}(r,p_t(r)) / dr] = partial_{H(p_r=0)}/partial_r
#            + partial_{H(p_r=0)}/partial_{p_t} partial_{p_t}/partial_r.
#  Here,
#  * the expression for p_t is given by PN_p_t.py
#  * the expression for [dE_{\rm GW}/dt + dM/dt] is given by
#            PN_dE_GW_dt_and_dM_dt.py
#    + Since [dE_{\rm GW}/dt + dM/dt] is a function of MOmega,
#        we also need input from the PN_MOmega.py Python module.

# Step 1: Construct full Hamiltonian
#         expression for a binary instantaneously
#         orbiting on the xy plane, store
#         result to Htot_xyplane_binary
def f_Htot_xyplane_binary(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r):
    def make_replacements(expr):
        zero = sp.sympify(0)
        one = sp.sympify(1)
        return (
            expr.subs(p1U[1], Pt)
            .subs(p2U[1], -Pt)
            .subs(p1U[2], zero)
            .subs(p2U[2], zero)
            .subs(p1U[0], -Pr)
            .subs(p2U[0], Pr)
            .subs(nU[0], one)
            .subs(nU[1], zero)
            .subs(nU[2], zero)
        )

    from . import PN_Hamiltonian_NS as H_NS

    H_NS.f_H_Newt__H_NS_1PN__H_NS_2PN(m1, m2, p1U, n12U, r)
    H_NS.f_H_NS_3PN(m1, m2, p1U, n12U, r)

    global Htot_xyplane_binary
    Htot_xyplane_binary = make_replacements(
        +H_NS.H_Newt + H_NS.H_NS_1PN + H_NS.H_NS_2PN + H_NS.H_NS_3PN
    )

    from . import PN_Hamiltonian_SO as H_SO

    H_SO.f_H_SO_1p5PN(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r)
    H_SO.f_H_SO_2p5PN(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r)
    H_SO.f_H_SO_3p5PN(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r)
    Htot_xyplane_binary += make_replacements(
        +H_SO.H_SO_1p5PN + H_SO.H_SO_2p5PN + H_SO.H_SO_3p5PN
    )

    from . import PN_Hamiltonian_SS as H_SS

    H_SS.f_H_SS_2PN(m1, m2, S1U, S2U, nU, r)
    H_SS.f_H_SS_S1S2_3PN(m1, m2, n12U, S1U, S2U, p1U, p2U, r)
    H_SS.f_H_SS_S1sq_S2sq_3PN(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r)
    Htot_xyplane_binary += make_replacements(
        +H_SS.H_SS_2PN + H_SS.H_SS_S1S2_3PN + H_SS.H_SS_S1sq_S2sq_3PN
    )

    from . import PN_Hamiltonian_SSS as H_SSS

    H_SSS.f_H_SSS_3PN(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r)
    Htot_xyplane_binary += make_replacements(+H_SSS.H_SSS_3PN)


# Function for computing dr/dt
def f_dr_dt(Htot_xyplane_binary, m1, m2, n12U, chi1U, chi2U, S1U, S2U, r):
    # First compute p_t
    from . import PN_p_t as pt

    pt.f_p_t(m1, m2, chi1U, chi2U, r)

    # Then compute dH_{circ}/dr = partial_H(p_r=0)/partial_r
    #                                  + partial_H(p_r=0)/partial_{p_t} partial_{p_t}/partial_r
    dHcirc_dr = +sp.diff(
        Htot_xyplane_binary.subs(Pr, sp.sympify(0)), r
    ) + sp.diff(Htot_xyplane_binary.subs(Pr, sp.sympify(0)), Pt) * sp.diff(
        pt.p_t, r
    )

    # Then compute M\Omega
    from . import PN_MOmega as MOm

    MOm.f_MOmega(m1, m2, chi1U, chi2U, r)

    # Next compute dE_GW_dt_plus_dM_dt
    from . import PN_dE_GW_dt_and_dM_dt as dEdt

    dEdt.f_dE_GW_dt_and_dM_dt(MOm.MOmega, m1, m2, n12U, S1U, S2U)

    # Finally, compute dr/dt
    global dr_dt
    dr_dt = dEdt.dE_GW_dt_plus_dM_dt / dHcirc_dr


# Next we compute p_r as a function of dr_dt (unknown) and known quantities using
# p_r \approx [dr/dt - (partial_H/partial_{p_r})|_{p_r=0}] * [(partial^2_{H}/partial_{p_r^2})|_{p_r=0}]^{-1}
def f_p_r_fullHam(m1, m2, n12U, n21U, chi1U, chi2U, S1U, S2U, p1U, p2U, r):
    f_Htot_xyplane_binary(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r)
    f_dr_dt(Htot_xyplane_binary, m1, m2, n12U, chi1U, chi2U, S1U, S2U, r)

    dHdpr_przero = sp.diff(Htot_xyplane_binary, Pr).subs(Pr, sp.sympify(0))
    d2Hdpr2_przero = sp.diff(sp.diff(Htot_xyplane_binary, Pr), Pr).subs(
        Pr, sp.sympify(0)
    )
    global p_r_fullHam
    p_r_fullHam = (dr_dt - dHdpr_przero) / (d2Hdpr2_przero)


# Here's the Ramos-Buades, Husa, and Pratten (2018)
#  approach for computing p_r.
# Transcribed from Eq 2.18 of
# Ramos-Buades, Husa, and Pratten (2018),
#   https://arxiv.org/abs/1810.00036
def f_p_r(m1, m2, n12U, n21U, chi1U, chi2U, S1U, S2U, p1U, p2U, r):
    q = m2 / m1  # It is assumed that q >= 1, so m2 >= m1.
    f_Htot_xyplane_binary(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r)
    f_dr_dt(Htot_xyplane_binary, m1, m2, n12U, chi1U, chi2U, S1U, S2U, r)
    chi1x = chi1U[0]
    chi1y = chi1U[1]
    chi1z = chi1U[2]
    chi2x = chi2U[0]
    chi2y = chi2U[1]
    chi2z = chi2U[2]
    p_r_num = (
        -dr_dt
        + (
            -(6 * q + 13) * q ** 2 * chi1x * chi2y / (4 * (q + 1) ** 4)
            - (6 * q + 1) * q ** 2 * chi2x * chi2y / (4 * (q + 1) ** 4)
            + chi1y
            * (
                -q * (q + 6) * chi1x / (4 * (q + 1) ** 4)
                - q * (13 * q + 6) * chi2x / (4 * (q + 1) ** 4)
            )
        )
        / r ** div(7, 2)
        + (
            +chi1z
            * (
                +3 * q * (5 * q + 2) * chi1x * chi2y / (2 * (q + 1) ** 4)
                - 3 * q ** 2 * (2 * q + 5) * chi2x * chi2y / (2 * (q + 1) ** 4)
            )
            + chi1y
            * chi2z
            * (
                +3 * q ** 2 * (2 * q + 5) * chi2x / (2 * (q + 1) ** 4)
                - 3 * q * (5 * q + 2) * chi1x / (2 * (q + 1) ** 4)
            )
        )
        / r ** 4
    )
    p_r_den = (
        -((q + 1) ** 2) / q
        - (-7 * q ** 2 - 15 * q - 7) / (2 * q * r)
        - (47 * q ** 4 + 229 * q ** 3 + 363 * q ** 2 + 229 * q + 47)
        / (8 * q * (q + 1) ** 2 * r ** 2)
        - (
            +(4 * q ** 2 + 11 * q + 12) * chi1z / (4 * q * (q + 1))
            + (12 * q ** 2 + 11 * q + 4) * chi2z / (4 * (q + 1))
        )
        / r ** div(5, 2)
        - (
            +(
                -53 * q ** 5
                - 357 * q ** 4
                - 1097 * q ** 3
                - 1486 * q ** 2
                - 842 * q
                - 144
            )
            * chi1z
            / (16 * q * (q + 1) ** 4)
            + (
                -144 * q ** 5
                - 842 * q ** 4
                - 1486 * q ** 3
                - 1097 * q ** 2
                - 357 * q
                - 53
            )
            * chi2z
            / (16 * (q + 1) ** 4)
        )
        / r ** div(7, 2)
        - (
            +(q ** 2 + 9 * q + 9) * chi1x ** 2 / (2 * q * (q + 1) ** 2)
            + (3 * q ** 2 + 5 * q + 3) * chi2x * chi1x / ((q + 1) ** 2)
            + (3 * q ** 2 + 8 * q + 3) * chi1y * chi2y / (2 * (q + 1) ** 2)
            - 9 * q ** 2 * chi2y ** 2 / (4 * (q + 1))
            + (3 * q ** 2 + 8 * q + 3) * chi1z * chi2z / (2 * (q + 1) ** 2)
            - 9 * q ** 2 * chi2z ** 2 / (4 * (q + 1))
            + (9 * q ** 3 + 9 * q ** 2 + q) * chi2x ** 2 / (2 * (q + 1) ** 2)
            + (
                -363 * q ** 6
                - 2608 * q ** 5
                - 7324 * q ** 4
                - 10161 * q ** 3
                - 7324 * q ** 2
                - 2608 * q
                - 363
            )
            / (48 * q * (q + 1) ** 4)
            - 9 * chi1y ** 2 / (4 * q * (q + 1))
            - 9 * chi1z ** 2 / (4 * q * (q + 1))
            - sp.pi ** 2 / 16
        )
        / r ** 3
    )
    global p_r
    p_r = p_r_num / p_r_den



# BSD 2-Clause License

# Copyright (c) 2018, Zachariah Etienne All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from sympy import N

from .NRPyPN_shortcuts import eval__P_t__and__P_r


def compute_quasicircular_momenta(
    mass_ratio, coordinate_distance, chi_plus, chi_minus
):
    """Wrapper around :py:func:`~eval__P_t__and__P_r does`. It is guaranteed that a number is returned

        # TODO: Clean up this docstring

    :param mass_ratio: Mass ratio of the binary (>= 1).
    :type mass_ratio: float
    :param coordinate_distance: Initial coordinate separation.
    :type coordinate_distance: float

    :param chi_plus: Dimensionless spin of the black hole on the positive side of the x
                     (or z) axis along the three directions.
    :type chi_plus: tuple/list with three numbers

    :param chi_minus: Dimensionless spin of the black hole on the negative side of the x
                     (or z) axis along the three directions.
    :type chi_minus: tuple/list with three numbers


    # The numerical values for
    #  * P_t: the tangential momentum
    #  * P_r: the radial momentum
    """
    (nchi1x, nchi1y, nchi1z), (nchi2x, nchi2y, nchi2z) = chi_plus, chi_minus

    return tuple(
        map(
            N,
            eval__P_t__and__P_r(
                mass_ratio,
                coordinate_distance,
                nchi1x,
                nchi1y,
                nchi1z,
                nchi2x,
                nchi2y,
                nchi2z,
            ),
        )
    )



# BSD 2-Clause License

# Copyright (c) 2018, Zachariah Etienne All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# As documented in the NRPyPN notebook
# NRPyPN_shortcuts.ipynb, this Python script
# provides useful shortcuts for inputting
# post-Newtonian expressions into SymPy/NRPy+

# Basic functions:
# dot(a,b): 3-vector dot product
# cross(a,b): 3-vector cross product
# div(a,b): a shortcut for SymPy's sp.Rational(a,b), to declare rational numbers
# num_eval(expr): Numerically evaluates NRPyPN expressions

# Author:  Zach Etienne
#          zachetie **at** gmail **dot* com

# Step 0: Add NRPy's directory to the path
# https://stackoverflow.com/questions/16780014/import-file-from-parent-directory
import sympy as sp  # SymPy: The Python computer algebra package upon which NRPy+ depends

from . import (
    indexedexpNRPyPN as ixp,  # NRPy+: Symbolic indexed expression (e.g., tensors, vectors, etc.) support
)

# Step 1: Declare several global variables used
#         throughout NRPyPN
m1, m2 = sp.symbols("m1 m2", real=True)
S1U = ixp.declarerank1("S1U")
S2U = ixp.declarerank1("S2U")
pU = ixp.declarerank1("pU")
nU = ixp.declarerank1("nU")

drdt = sp.symbols("drdt", real=True)
Pt, Pr = sp.symbols("Pt Pr", real=True)
# Some references use r, others use q to represent the
#   distance between the two point masses. This is rather
#   confusing since q is also used to represent the
#   mass ratio m2/m1. However, q is the canonical position
#   variable name in Hamiltonian mechanics, so both are
#   well justified. It should be obvious which is which
#   throughout NRPyPN.
r, q = sp.symbols("r q", real=True)
chi1U = ixp.declarerank1("chi1U")
chi2U = ixp.declarerank1("chi2U")

# Euler-Mascheroni gamma constant:
gamma_EulerMascheroni = sp.EulerGamma

# Derived quantities used in Damour et al papers:
n12U = ixp.zerorank1()
n21U = ixp.zerorank1()
p1U = ixp.zerorank1()
p2U = ixp.zerorank1()
for i in range(3):
    n12U[i] = +nU[i]
    n21U[i] = -nU[i]
    p1U[i] = +pU[i]
    p2U[i] = -pU[i]

# Step 2.a: Define dot and cross product of vectors
def dot(vec1, vec2):
    vec1_dot_vec2 = sp.sympify(0)
    for i in range(3):
        vec1_dot_vec2 += vec1[i] * vec2[i]
    return vec1_dot_vec2


def cross(vec1, vec2):
    vec1_cross_vec2 = ixp.zerorank1()
    LeviCivitaSymbol = ixp.LeviCivitaSymbol_dim3_rank3()
    for i in range(3):
        for j in range(3):
            for k in range(3):
                vec1_cross_vec2[i] += (
                    LeviCivitaSymbol[i][j][k] * vec1[j] * vec2[k]
                )
    return vec1_cross_vec2


# Step 2.b: Construct rational numbers a/b via div(a,b)
def div(a, b):
    return sp.Rational(a, b)


# Step 3: num_eval(expr), a means to numerically evaluate SymPy/NRPyPN
#         expressions
def num_eval(
    expr,
    qmassratio=1.0,  # must be >= 1
    nr=12.0,  # Orbital separation
    nchi1x=+0.0,
    nchi1y=+0.0,
    nchi1z=+0.0,
    nchi2x=+0.0,
    nchi2y=+0.0,
    nchi2z=+0.0,
    nPt=None,
    ndrdt=None,
):

    # DERIVED QUANTITIES BELOW
    # We want m1+m2 = 1, so that
    #         m2/m1 = qmassratio
    # We find below:
    nm1 = 1 / (1 + qmassratio)
    nm2 = qmassratio / (1 + qmassratio)
    # This way nm1+nm2 = (qmassratio+1)/(1+qmassratio) = 1 CHECK
    #      and nm2/nm1 = qmassratio                        CHECK

    nS1U0 = nchi1x * nm1 ** 2
    nS1U1 = nchi1y * nm1 ** 2
    nS1U2 = nchi1z * nm1 ** 2
    nS2U0 = nchi2x * nm2 ** 2
    nS2U1 = nchi2y * nm2 ** 2
    nS2U2 = nchi2z * nm2 ** 2

    if nPt != None:
        expr2 = expr.subs(Pt, nPt)
        expr = expr2
    if ndrdt != None:
        expr2 = expr.subs(drdt, ndrdt)
        expr = expr2
    return (
        expr.subs(m1, nm1)
        .subs(m2, nm2)
        .subs(S1U[0], nS1U0)
        .subs(S1U[1], nS1U1)
        .subs(S1U[2], nS1U2)
        .subs(S2U[0], nS2U0)
        .subs(S2U[1], nS2U1)
        .subs(S2U[2], nS2U2)
        .subs(chi1U[0], nchi1x)
        .subs(chi1U[1], nchi1y)
        .subs(chi1U[2], nchi1z)
        .subs(chi2U[0], nchi2x)
        .subs(chi2U[1], nchi2y)
        .subs(chi2U[2], nchi2z)
        .subs(r, nr)
        .subs(q, nr)
        .subs(sp.pi, sp.N(sp.pi))
        .subs(gamma_EulerMascheroni, sp.N(sp.EulerGamma))
    )


# Shortcut function to just evaluate P_t and P_r
#  -= Inputs =-
#  * qmassratio: mass ratio q>=1
#  * nr: Orbital separation (total coordinate distance between black holes)
#  * nchi1x,nchi1y,nchi1z: dimensionless spin vector for BH 1
#  * nchi2x,nchi2y,nchi2z: dimensionless spin vector for BH 2
#  -= Outputs =-
# The numerical values for
#  * P_t: the tangential momentum
#  * P_r: the radial momentum
def eval__P_t__and__P_r(
    qmassratio, nr, nchi1x, nchi1y, nchi1z, nchi2x, nchi2y, nchi2z
):

    # Compute p_t, the tangential component of momentum
    from . import PN_p_t as pt

    pt.f_p_t(m1, m2, chi1U, chi2U, r)

    # Compute p_r, the radial component of momentum
    from . import PN_p_r as pr

    pr.f_p_r(m1, m2, n12U, n21U, chi1U, chi2U, S1U, S2U, p1U, p2U, r)

    nPt = num_eval(
        pt.p_t,
        qmassratio=qmassratio,
        nr=nr,
        nchi1x=nchi1x,
        nchi1y=nchi1y,
        nchi1z=nchi1z,
        nchi2x=nchi2x,
        nchi2y=nchi2y,
        nchi2z=nchi2z,
    )

    nPr = num_eval(
        pr.p_r,
        qmassratio=qmassratio,
        nr=nr,
        nchi1x=nchi1x,
        nchi1y=nchi1y,
        nchi1z=nchi1z,
        nchi2x=nchi2x,
        nchi2y=nchi2y,
        nchi2z=nchi2z,
        nPt=nPt,
    )
    return nPt, nPr



# BSD 2-Clause License

# Copyright (c) 2018, Zachariah Etienne All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# As documented in the NRPyPN notebook
# PN-Hamiltonian-SSS.ipynb, this Python script
# generates spin-spin-spin coupling pieces of the
# post-Newtonian (PN) Hamiltonian, up to and
# including 3PN order.

# Core functions:
# f_H_SSS_3PN(m1,m2, n12U,n21U, S1U,S2U, p1U,p2U, q)
#       Compute the complete H_SSS_3PN term and store to
#                     global variable of the same name.

# Author:  Zach Etienne
#          zachetie **at** gmail **dot* com

# Step 0: Add NRPy's directory to the path
# https://stackoverflow.com/questions/16780014/import-file-from-parent-directory
from . import (
    indexedexpNRPyPN as ixp,  # NRPy+: Symbolic indexed expression (e.g., tensors, vectors, etc.) support
)
from .NRPyPN_shortcuts import (  # NRPyPN: shortcuts for e.g., vector operations
    cross,
    div,
    dot,
)


#################################
#################################
# Step 1: 3PN spin-spin-spin term, from Eq. 3.12 of
#        Levi and Steinhoff (2015):
#     https://arxiv.org/abs/1410.2601
def f_H_SSS_3PN(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r12):
    def f_H_SSS_3PN_pt(m1, m2, nU, S1U, S2U, p1U, p2U, r):
        p2_minus_m2_over_4m1_p1 = ixp.zerorank1()
        for i in range(3):
            p2_minus_m2_over_4m1_p1[i] = p2U[i] - m2 / (4 * m1) * p1U[i]
        H_SSS_3PN_pt = (
            +div(3, 2)
            * (
                +dot(S1U, S1U) * dot(S2U, cross(nU, p1U))
                + dot(S1U, nU) * dot(S2U, cross(S1U, p1U))
                - 5 * dot(S1U, nU) ** 2 * dot(S2U, cross(nU, p1U))
                + dot(nU, cross(S1U, S2U))
                * (+dot(S1U, p1U) - 5 * dot(S1U, nU) * dot(p1U, nU))
            )
            - 3
            * m1
            / (2 * m2)
            * (
                +dot(S1U, S1U) * dot(S2U, cross(nU, p2U))
                + 2 * dot(S1U, nU) * dot(S2U, cross(S1U, p2U))
                - 5 * dot(S1U, nU) ** 2 * dot(S2U, cross(nU, p2U))
            )
            - dot(cross(S1U, nU), p2_minus_m2_over_4m1_p1)
            * (dot(S1U, S1U) - 5 * dot(S1U, nU) ** 2)
        ) / (m1 ** 2 * r ** 4)
        return H_SSS_3PN_pt

    global H_SSS_3PN
    H_SSS_3PN = +f_H_SSS_3PN_pt(
        m1, m2, n12U, S1U, S2U, p1U, p2U, r12
    ) + f_H_SSS_3PN_pt(m2, m1, n21U, S2U, S1U, p2U, p1U, r12)



# BSD 2-Clause License

# Copyright (c) 2018, Zachariah Etienne All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# As documented in the NRPyPN notebook
# PN-dE_GW_dt.ipynb, this Python script
# generates dE_GW/dt at highest known
# post-Newtonian order (as of 2015, at
# least).

# Core functions:
# dE_GW_dt_OBKPSS2015_consts(m1,m2, n12U, S1U,S2U):
#       Define constants used in the dE_GW/dt expression.
# f_dE_GW_dt(mOmega, m1,m2, n12U, S1U,S2U):
#       Compute dE_GW_dt and store to global variable of the same name.

# Author:  Zach Etienne
#          zachetie **at** gmail **dot* com

# Step 0: Add NRPy's directory to the path
# https://stackoverflow.com/questions/16780014/import-file-from-parent-directory
import sys  # Standard Python modules for multiplatform OS-level functions

import sympy as sp  # SymPy: The Python computer algebra package upon which NRPy+ depends

from . import (
    indexedexpNRPyPN as ixp,  # NRPy+: Symbolic indexed expression (e.g., tensors, vectors, etc.) support
)
from .NRPyPN_shortcuts import (  # NRPyPN: shortcuts for e.g., vector operations
    div,
    dot,
    gamma_EulerMascheroni,
)

#################################
#################################

# Constants given in Eqs A1-13 of https://arxiv.org/abs/1502.01747
def dE_GW_dt_OBKPSS2015_consts(m1, m2, _n12U, S1U, S2U):  # _n12U unused.
    # define scalars:
    m = m1 + m2
    nu = m1 * m2 / m ** 2
    delta = (m1 - m2) / m
    # define vectors:
    Stot = ixp.zerorank1()
    Sigma = ixp.zerorank1()
    l = ixp.zerorank1()
    l[2] = sp.sympify(1)
    chi1U = ixp.zerorank1()
    chi2U = ixp.zerorank1()
    chi_s = ixp.zerorank1()
    chi_a = ixp.zerorank1()
    for i in range(3):
        Stot[i] = S1U[i] + S2U[i]
        Sigma[i] = (m1 + m2) / m2 * S2U[i] - (m1 + m2) / m1 * S1U[i]
        chi1U[i] = S1U[i] / m1 ** 2
        chi2U[i] = S2U[i] / m2 ** 2
        chi_s[i] = div(1, 2) * (chi1U[i] + chi2U[i])
        chi_a[i] = div(1, 2) * (chi1U[i] - chi2U[i])
    # define scalars that depend on vectors
    s_l = dot(Stot, l) / m ** 2
    # s_n = dot(Stot,n12U)/m**2
    sigma_l = dot(Sigma, l) / m ** 2
    # sigma_n = dot(Sigma,n12U)/m**2
    return nu, delta, l, chi_a, chi_s, s_l, sigma_l


#################################
#################################

# Based on Eqs A22-28 of https://arxiv.org/abs/1502.01747, with
#       Eq A.14 of https://arxiv.org/abs/0709.0093 for Mdot
#       and correction on b[7] term by comparison with
#  https://link.springer.com/content/pdf/10.12942/lrr-2014-2.pdf
def f_dE_GW_dt_and_dM_dt(mOmega, m1, m2, n12U, S1U, S2U):
    def f_compute_quantities(mOmega, m1, m2, n12U, S1U, S2U, which_quantity):
        if not which_quantity in ("dM_dt", "dE_GW_dt", "dE_GW_dt_plus_dM_dt"):
            print(
                "which_quantity == " + str(which_quantity) + " not supported!"
            )
            sys.exit(1)

        nu, delta, l, chi_a, chi_s, s_l, sigma_l = dE_GW_dt_OBKPSS2015_consts(
            m1, m2, n12U, S1U, S2U
        )
        x = (mOmega) ** div(2, 3)

        # Compute b_5_Mdot:
        b_5_Mdot = -div(1, 4) * (
            +(1 - 3 * nu)
            * dot(chi_s, l)
            * (1 + 3 * dot(chi_s, l) ** 2 + 9 * dot(chi_a, l) ** 2)
            + (1 - nu)
            * delta
            * dot(chi_a, l)
            * (1 + 3 * dot(chi_a, l) ** 2 + 9 * dot(chi_s, l) ** 2)
        )
        if which_quantity == "dM_dt":
            return div(32, 5) * nu ** 2 * x ** 5 * b_5_Mdot * x ** div(5, 2)

        b = ixp.zerorank1(DIM=10)
        b[2] = -div(1247, 336) - div(35, 12) * nu
        b[3] = +4 * sp.pi - 4 * s_l - div(5, 4) * delta * sigma_l
        b[4] = (
            -div(44711, 9072)
            + div(9271, 504) * nu
            + div(65, 18) * nu ** 2
            + (+div(287, 96) + div(1, 24) * nu) * dot(chi_s, l) ** 2
            - (+div(89, 96) + div(7, 24) * nu) * dot(chi_s, chi_s)
            + (+div(287, 96) - 12 * nu) * dot(chi_a, l) ** 2
            + (-div(89, 96) + 4 * nu) * dot(chi_a, chi_a)
            + div(287, 48) * delta * dot(chi_s, l) * dot(chi_a, l)
            - div(89, 48) * delta * dot(chi_s, chi_a)
        )
        b[5] = (
            -div(8191, 672) * sp.pi
            - div(9, 2) * s_l
            - div(13, 16) * delta * sigma_l
            + nu
            * (
                -div(583, 24) * sp.pi
                + div(272, 9) * s_l
                + div(43, 4) * delta * sigma_l
            )
        )
        if which_quantity == "dE_GW_dt_plus_dM_dt":
            b[5] += b_5_Mdot
        b[6] = (
            +div(6643739519, 69854400)
            + div(16, 3) * sp.pi ** 2
            - div(1712, 105) * gamma_EulerMascheroni
            - div(856, 105) * sp.log(16 * x)
            + (-div(134543, 7776) + div(41, 48) * sp.pi ** 2) * nu
            - div(94403, 3024) * nu ** 2
            - div(775, 324) * nu ** 3
            - 16 * sp.pi * s_l
            - div(31, 6) * sp.pi * delta * sigma_l
        )
        b[7] = (
            +(
                +div(476645, 6804)
                + div(6172, 189) * nu
                - div(2810, 27) * nu ** 2
            )
            * s_l
            + (+div(9535, 336) + div(1849, 126) * nu - div(1501, 36) * nu ** 2)
            * delta
            * sigma_l
            + (
                -div(16285, 504)
                + div(214745, 1728) * nu
                + div(193385, 3024) * nu ** 2
            )
            * sp.pi
        )
        b[8] = (
            +(-div(3485, 96) * sp.pi + div(13879, 72) * sp.pi * nu) * s_l
            + (-div(7163, 672) * sp.pi + div(130583, 2016) * sp.pi * nu)
            * delta
            * sigma_l
        )
        b_sum = sp.sympify(1)
        for k in range(9):
            b_sum += b[k] * x ** div(k, 2)
        return div(32, 5) * nu ** 2 * x ** 5 * b_sum

    global dE_GW_dt_plus_dM_dt, dE_GW_dt, dM_dt
    dE_GW_dt_plus_dM_dt = f_compute_quantities(
        mOmega, m1, m2, n12U, S1U, S2U, "dE_GW_dt_plus_dM_dt"
    )
    dE_GW_dt = f_compute_quantities(mOmega, m1, m2, n12U, S1U, S2U, "dE_GW_dt")
    dM_dt = f_compute_quantities(mOmega, m1, m2, n12U, S1U, S2U, "dM_dt")



#!/usr/bin/env python3

# Copyright (C) 2020-2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.simfactory` module provides functions to interface with
SimFactory.

The available functions are:

- :py:func:`~.simfactory_option`, which helps setting up options that are
controllable by SimFactory.

- :py:func:`~.write_parfile`, which write the `.par` file from a template
when the file is executed as `.rpar` by SimFactory.

"""
import inspect  # Needed to capture locals() of the caller to write the parfile
import re
import sys

from jhuki.parfile import write_one_parfile_from_template


def simfactory_option(name, default):
    """Set a variable to a SimFactory-controllable parameter.

    SimFactory can define arbitrary parameters using the ``--define`` keyword.
    With this function, you can define a variable that takes a specified default
    value, unless the keyword is used.

    For example:

    .. code-block::python

        par_b = simfactory_option("@PARB@", 6.5)

    par_b is a Python variable with the value 6.5 unless ``--define PARB=XX`` is
    passed to SimFactory. The way this works is that SimFactory performs a
    substitution of objects with the double @ symbols. For this reason, the
    name parameter to this function has to be surrounded by @.

    We convert all the numbers to floats, so if you need integers, you have to
    explicitly cast the variable to int, for example:

    .. code-block::python

        num_ref_radii = int(simfactory_option("@REFLEVELS@", 10))

    :param name: Name of the SimFactory define key. It has to be with surrounded
                 by @, for example, @XMAX@.
    :type name: str
    :param default: Default value that the variable has to have if define is not
                    specified.
    :type default: int, float, or str

    :returns: The default value if the define key is not passed to SimFactory,
              otherwise the value defined with SimFactory.
    :rtype: float or str (integers are converted)

    """
    if not isinstance(name, str):
        raise TypeError(f"name has to a string, it is {name}")

    # If default is a string, we will keep it a string. If it is an int, we
    # convert it to float. Often one sets parameter like par_b = 6, but here 6
    # has to be intended as a float, or it can cause troubles for other
    # calculations.
    type_ = float if isinstance(default, int) else type(default)

    return type_(name) if name[0] != "@" else type_(default)


def write_parfile(template: str, file_name=None) -> None:
    """Write parfile from given template using locals() as substitution
    dictionary.

    The variables in template get substituted with the values in the local scope
    of the calling function. So, you can safely put this function at the end of
    your rpar file to have the template substituted with the variables you have
    defined there.

    :param template: Template of parfile to write. See
                     :py:func:`~.write_one_parfile_from_template`
    :type template: str
    :param file_name: Name of the file that is calling this function.
                      (Used only for debug purposes.)
    :type file_name: str

    """
    # Compact notation to say: if file_name is not None, file_name = file_name,
    # otherwise it is sys.argv[0]
    file_name = file_name or sys.argv[0]

    # The regular expression matches something that has the form FILENAME.rpar
    rpar_rx = re.compile(r"^(.*)\.rpar$")

    if not rpar_rx.match(file_name):
        raise RuntimeError("Function was not called by a rpar file")

    # Replace "file_name.rpar" with "file_name.par"
    output_file_name = rpar_rx.sub(r"\1.par", file_name)

    # We want to capture the locals() of the caller (to use it as substitution
    # dictionary). This function should be called directly from the `rpar` file

    # From https://stackoverflow.com/a/62885157
    frame = inspect.currentframe().f_back
    try:
        # frame.f_locals is the locals() of the calling function
        subs = frame.f_locals

        # From here, we have to remove the template variable itself (otherwise
        # we would have a template-ception)
        #
        # The easiest way to do this is to match the value in the dictionary
        # subs. This is not particularly efficient, but we expect subs to be
        # small
        subs_no_template = {k: v for k, v in subs.items() if v != template}

        write_one_parfile_from_template(
            template, sub_dict=subs_no_template, out_file=output_file_name
        )
    finally:
        # Ensure that we release the resources
        del frame



#!/usr/bin/env python3

# Copyright (C) 2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.twopunctures` module provides infrastructure to prepare
binary black simulations.

Some functions are provided:
- :py:func:`~.prepare_quasicircular_inspiral`, which returns a :py:class:`~.TwoPunctures`
  with parameters set up to lead to a quasi-circular inspiral. The linear momenta
  required for this are computed with a 3.5PN expansion by NRPyPN.

"""

from functools import lru_cache

from jhuki.base import BaseThorn
from jhuki.externals.nrpypn import compute_quasicircular_momenta


def prepare_quasicircular_inspiral(
    mass_ratio,
    coordinate_distance,
    total_bare_mass=1,
    chi_plus=(0, 0, 0),
    chi_minus=(0, 0, 0),
    **kwargs,
):
    """Return a :py:class:`~.TwoPunctures` that describes a quasi-circular inspiral.
    We always assume that the plus puncture is the most massive one.

    :param mass_ratio: Mass ratio of the binary.
    :type mass_ratio: float
    :param coordinate_distance: Initial coordinate separation.
    :type coordinate_distance: float
    :param total_bare_mass: Rescale masses and momenta so that the total bare mass of the
                            system is this.
    :type total_bare_mass: float

    :param chi_plus: Dimensionless spin of the black hole on the positive side of the x
                     (or z) axis along the three directions.
    :type chi_plus: tuple/list with three numbers

    :param chi_minus: Dimensionless spin of the black hole on the negative side of the x
                     (or z) axis along the three directions.
    :type chi_minus: tuple/list with three numbers

    Unknown arguments are passed to :py:class:`~.TwoPunctures`.

    :returns: A :py:class:`~.TwoPunctures` for a quasi-circular inspiral.
    :rtype: :py:class:`~.TwoPunctures`
    """

    if mass_ratio < 1:
        mass_ratio = 1 / mass_ratio

    Pt, Pr = compute_quasicircular_momenta(
        mass_ratio, coordinate_distance, chi_plus, chi_minus
    )

    Pt, Pr = Pt * total_bare_mass, Pr * total_bare_mass

    momenta_plus = (-Pr, Pt, 0)
    momenta_minus = (Pr, -Pt, 0)

    mass_plus = mass_ratio / (1 + mass_ratio) * total_bare_mass
    mass_minus = 1 / (1 + mass_ratio) * total_bare_mass

    return TwoPunctures(
        mass_plus,
        mass_minus,
        coordinate_distance,
        momenta_plus=momenta_plus,
        momenta_minus=momenta_minus,
        chi_plus=chi_plus,
        chi_minus=chi_minus,
        give_bare_mass=False,
        **kwargs,
    )


class TwoPunctures(BaseThorn):
    """The :py:class:`~.TwoPunctures` class contains all the information related to
    one or two black holes, their locations, and properties.

    This class can set up parameters for the ``TwoPunctures`` thorn making sure that
    the center of mass is in the center of the grid (unless differently specified).

    This class is immutable.

    :ivar mass_plus: Mass of the black hole on the positive side of the x (or z) axis.
    :vartype mass_plus: float
    :ivar mass_minus: Mass of the black hole on the negative side of the x (or z) axis.
    :vartype mass_minus: float
    :ivar coordinate_separation: Initial coordinate separation.
    :vartype coordinate_separation: float

    :ivar coord_x_plus: Initial coordinate location of the center of the black hole on the
                        positive side of the x (or z) axis.
    :vartype coord_x_plus: float
    :ivar coord_x_minus: Initial coordinate location of the center of the black hole on the
                         negative side of the x (or z) axis.
    :vartype coord_x_minus: float

    :ivar momenta_plus: Array with the linear momenta along the three directions for the
                        black hole on the positive side of the x (or z) axis.
    :vartype momenta_plus: list/tuple
    :ivar momenta_minus: Array with the linear momenta along the three directions for the
                         black hole on the positive side of the x (or z) axis.
    :vartype momenta_minus: list/tuple

    :ivar chi_plus: Array with the dimensionless_spin along the three directions for the
                    black hole on the positive side of the x (or z) axis.
    :vartype chi_plus: list/tuple
    :ivar chi_minus: Array with the dimensionless spin along the three directions for the
                     black hole on the positive side of the x (or z) axis.
    :vartype chi_minus: list/tuple

    :ivar S_plus: Array with the angular momentum along the three directions for the
                  black hole on the positive side of the x (or z) axis.
    :vartype S_plus: list/tuple
    :ivar S_minus: Array with the angular momentum along the three directions for the
                   black hole on the positive side of the x (or z) axis.
    :vartype S_minus: list/tuple

    :ivar center_offset: Move the geometric center of the system by this amount.
    :vartype center_offset: list/tuple

    :ivar swap_xz: If True, activate the ``swap_xz`` parameter in TwoPunctures.
    :vartype swap_xz: bool

    :ivar give_bare_mass: If True, set this parameter to True in the parfile.
    :vartype give_bare_mass: bool

    :ivar initial_alpha: Prescription to use for the initial lapse.
    :vartype initial_alpha: str
    """

    def __init__(
        self,
        mass_plus,
        mass_minus,
        coordinate_distance,
        momenta_plus=None,
        momenta_minus=None,
        chi_plus=None,
        chi_minus=None,
        swap_xz=False,
        give_bare_mass=False,
        initial_alpha="twopunctures-averaged",
    ):
        """Constructor.

        :param mass_plus: Mass of the black hole on the positive side of the x (or z) axis.
        :type mass_plus: float
        :param mass_minus: Mass of the black hole on the negative side of the x (or z) axis.
        :type mass_minus: float
        :param coordinate_separation: Initial coordinate separation.
        :type coordinate_separation: float
        :param par_b: Half of the initial separation.
        :type par_b: float

        :param momenta_plus: Array with the linear momenta along the three directions for the
                             black hole on the positive side of the x (or z) axis.
                             If ``swap_xz`` is True, these are the values after swapping.
                             For example ``(0, 0, 0.5)`` is a vector along the z axis.
                             (This is the opposite of what happens in ``TwoPunctures``, where
                              you have to define the values before swapping)
        :type momenta_plus: list/tuple
        :param momenta_minus: Array with the linear momenta along the three directions for the
                             black hole on the positive side of the x (or z) axis.
                             If ``swap_xz`` is True, these are the values after swapping.
                             For example ``(0, 0, 0.5)`` is a vector along the z axis.
                             (This is the opposite of what happens in ``TwoPunctures``, where
                              you have to define the values before swapping)
        :type momenta_minus: list/tuple

        :param chi_plus: Array with the dimensionless_spin along the three directions for the
                        black hole on the positive side of the x (or z) axis.
                             If ``swap_xz`` is True, these are the values after swapping.
                             For example ``(0, 0, 0.5)`` is a vector along the z axis.
                             (This is the opposite of what happens in ``TwoPunctures``, where
                              you have to define the values before swapping)
        :type chi_plus: list/tuple
        :param chi_minus: Array with the dimensionless spin along the three directions for the
                         black hole on the positive side of the x (or z) axis.
                             If ``swap_xz`` is True, these are the values after swapping.
                             For example ``(0, 0, 0.5)`` is a vector along the z axis.
                             (This is the opposite of what happens in ``TwoPunctures``, where
                              you have to define the values before swapping)
        :type chi_minus: list/tuple

        :param swap_xz: If True, activate the ``swap_xz`` parameter in TwoPunctures.
        :type swap_xz: bool

        :param give_bare_mass: If True, set this parameter to True in the parfile.
        :type give_bare_mass: bool

        :param initial_alpha: Prescription to use for the initial lapse.
        :type initial_alpha: str
        """

        # TODO: No sanity checks are performed here. We should at least check that P < m and so on

        def _par_or_zeros(par):
            # Return the parameter if it is not None, otherwise zeros
            return par if par is not None else (0.0, 0.0, 0.0)

        self.mass_plus = mass_plus
        self.mass_minus = mass_minus
        self.coordinate_distance = coordinate_distance

        self.par_b = coordinate_distance / 2

        total_mass = self.mass_plus + self.mass_minus

        self.coord_x_plus = (
            self.coordinate_distance * self.mass_minus / total_mass
        )
        self.coord_x_minus = (
            -self.coordinate_distance * self.mass_plus / total_mass
        )

        self.momenta_plus = _par_or_zeros(momenta_plus)
        self.momenta_minus = _par_or_zeros(momenta_minus)
        self.chi_plus = _par_or_zeros(chi_plus)
        self.chi_minus = _par_or_zeros(chi_minus)

        # Specific angular momentum
        self.a_plus = tuple(cp * self.mass_plus for cp in self.chi_plus)
        self.a_minus = tuple(cm * self.mass_minus for cm in self.chi_minus)

        # Angular momentum
        self.S_plus = tuple(cp * self.mass_plus**2 for cp in self.chi_plus)
        self.S_minus = tuple(
            cm * self.mass_minus**2 for cm in self.chi_minus
        )

        self.center_offset = (self.coord_x_plus - self.par_b, 0.0, 0.0)

        self.swap_xz = swap_xz
        self.give_bare_mass = give_bare_mass
        self.initial_alpha = initial_alpha

    @property
    @lru_cache(1)
    def parfile_code(self):
        """Return the code you would put in your parfile."""

        def assign_parameter(param, value, which_bh=None):
            if which_bh is None:
                return f"TwoPunctures::{param} = {value}"

            # which_bh is either _plus or _minus

            # If we have swap_xz, we must swap the values of the indices 0
            # and 2, so we prepare a simple look up table for the correct index:
            _swap = {
                0: (2 if self.swap_xz else 0),
                1: 1,
                2: (0 if self.swap_xz else 2),
            }
            return "\n".join(
                f"TwoPunctures::{param}{which_bh}[{index}] = {value[_swap[index]]}"
                for index in range(3)
            )

        ret = []

        ret.append(
            f"""\
ADMBase::initial_data = "twopunctures"
ADMBase::initial_lapse = "{self.initial_alpha}"
ADMBase::initial_shift = "zero"
ADMBase::initial_dtlapse = "zero"
ADMBase::initial_dtshift = "zero"

TwoPunctures::give_bare_mass = "{"yes" if self.give_bare_mass else "no"}"
TwoPunctures::par_b = {self.par_b}
TwoPunctures::target_m_plus = {self.mass_plus}
TwoPunctures::target_m_minus = {self.mass_minus}
TwoPunctures::par_m_plus = {self.mass_plus}
TwoPunctures::par_m_minus = {self.mass_minus}"""
        )

        ret.append(assign_parameter("par_P", self.momenta_plus, "_plus"))
        ret.append(assign_parameter("par_P", self.momenta_minus, "_minus"))
        ret.append(assign_parameter("par_S", self.S_plus, "_plus"))
        ret.append(assign_parameter("par_S", self.S_minus, "_minus"))
        ret.append(assign_parameter("center_offset", self.center_offset, ""))

        if self.swap_xz:
            ret.append(assign_parameter("swap_xz", "yes"))

        return "\n".join(ret)



# BSD 2-Clause License

# Copyright (c) 2018, Zachariah Etienne All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# As documented in the NRPyPN notebook
# PN-Hamiltonian-Nonspinning.ipynb, this Python script
# generates nonspinning pieces of the post-Newtonian (PN)
# Hamiltonian, up to and including third PN order.

# Basic functions:
# f_H_Newt__H_NS_1PN__H_NS_2PN(m1,m2, pU, nU, q): Compute H_Newt,
#                                                 H_NS_1PN, and H_NS_2PN
#                                                 and store to global
#                                                 variables of the same
#                                                 names.
# f_H_NS_3PN: Compute H_NS_3PN, and store to global variable of same name

# Author:  Zach Etienne
#          zachetie **at** gmail **dot* com

# Step 0: Add NRPy's directory to the path
# https://stackoverflow.com/questions/16780014/import-file-from-parent-directory
import sympy as sp  # SymPy: The Python computer algebra package upon which NRPy+ depends

from . import (
    indexedexpNRPyPN as ixp,  # NRPy+: Symbolic indexed expression (e.g., tensors, vectors, etc.) support
)
from .NRPyPN_shortcuts import (
    div,
    dot,
)  # NRPyPN: shortcuts for e.g., vector operations


def f_H_Newt__H_NS_1PN__H_NS_2PN(m1, m2, PU, nU, q):
    mu = m1 * m2 / (m1 + m2)
    eta = m1 * m2 / (m1 + m2) ** 2
    pU = ixp.zerorank1()
    for i in range(3):
        pU[i] = PU[i] / mu

    global H_Newt, H_NS_1PN, H_NS_2PN
    H_Newt = mu * (+div(1, 2) * dot(pU, pU) - 1 / q)

    H_NS_1PN = mu * (
        +div(1, 8) * (3 * eta - 1) * dot(pU, pU) ** 2
        - div(1, 2) * ((3 + eta) * dot(pU, pU) + eta * dot(nU, pU) ** 2) / q
        + div(1, 2) / q ** 2
    )

    H_NS_2PN = mu * (
        +div(1, 16) * (1 - 5 * eta + 5 * eta ** 2) * dot(pU, pU) ** 3
        + div(1, 8)
        * (
            +(5 - 20 * eta - 3 * eta ** 2) * dot(pU, pU) ** 2
            - 2 * eta ** 2 * dot(nU, pU) ** 2 * dot(pU, pU)
            - 3 * eta ** 2 * dot(nU, pU) ** 4
        )
        / q
        + div(1, 2)
        * ((5 + 8 * eta) * dot(pU, pU) + 3 * eta * dot(nU, pU) ** 2)
        / q ** 2
        - div(1, 4) * (1 + 3 * eta) / q ** 3
    )


def f_H_NS_3PN(m1, m2, PU, nU, q):
    mu = m1 * m2 / (m1 + m2)
    eta = m1 * m2 / (m1 + m2) ** 2
    pU = ixp.zerorank1()
    for i in range(3):
        pU[i] = PU[i] / mu

    global H_NS_3PN
    H_NS_3PN = mu * (
        +div(1, 128)
        * (-5 + 35 * eta - 70 * eta ** 2 + 35 * eta ** 3)
        * dot(pU, pU) ** 4
        + div(1, 16)
        * (
            +(-7 + 42 * eta - 53 * eta ** 2 - 5 * eta ** 3) * dot(pU, pU) ** 3
            + (2 - 3 * eta) * eta ** 2 * dot(nU, pU) ** 2 * dot(pU, pU) ** 2
            + 3 * (1 - eta) * eta ** 2 * dot(nU, pU) ** 4 * dot(pU, pU)
            - 5 * eta ** 3 * dot(nU, pU) ** 6
        )
        / q
        + (
            +div(1, 16) * (-27 + 136 * eta + 109 * eta ** 2) * dot(pU, pU) ** 2
            + div(1, 16)
            * (+17 + 30 * eta)
            * eta
            * dot(nU, pU) ** 2
            * dot(pU, pU)
            + div(1, 12) * (+5 + 43 * eta) * eta * dot(nU, pU) ** 4
        )
        / q ** 2
        + (
            +(
                -div(25, 8)
                + (div(1, 64) * sp.pi ** 2 - div(335, 48)) * eta
                - div(23, 8) * eta ** 2
            )
            * dot(pU, pU)
            + (-div(85, 16) - div(3, 64) * sp.pi ** 2 - div(7, 4) * eta)
            * eta
            * dot(nU, pU) ** 2
        )
        / q ** 3
        + (+div(1, 8) + (div(109, 12) - div(21, 32) * sp.pi ** 2) * eta)
        / q ** 4
    )



#!/usr/bin/env python3

# Copyright (C) 2020-2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.


"""The :py:mod:`~.parfile` module provides functions to programmatically write
parameter files. The main functions provided are:

- :py:func:`~.write_one_parfile_from_template`. This function takes a Python
  string template, a dictionary with the instructions to perform the
  substitution, and the name of the file where to write the output.
- :py:func:`~.write_many_parfiles_from_template`. This function calls the
  previous function (:py:func:`~.write_one_parfile_from_template`) multiple times
  on a multiple substitution lists to produce multiple files.

"""

import logging
from string import Template

from jhuki import __version__
from jhuki.base import BaseThorn


def write_one_parfile_from_template(template, sub_dict, out_file):
    """Write a parfile to out_file from template with substitution
    specified by the dictionary sub_dict.

    We add an header with the list of variables that were introduced.

    :param template: Basic template with parameters that have to be
                     substituted. It follows Python's template
                     conventions, so the variables to be substituted
                     are to be prefixed with a dollar sign.
    :type template: str
    :param sub_dict: Dictionary that maps entries that have to be
                     substituted with their value.
    :type sub_dict: dict
    :param out_file: Path of the output file
    :type out_file: str
    """

    # We add an header to save what variables were substituted
    substituted_variables = []
    for key, val in sorted(sub_dict.items()):
        # We ignore BaseThorns because they clutter the header with information
        # that is already available in the parfile. We also ignore functions.
        if not isinstance(val, BaseThorn) and not callable(val):
            # We cast everything to string and substitute all the newlines with
            # commented newlines so that we can ensure that multiline
            # parameters are properly commented.
            val = str(val).replace("\n", "\n# ")
            substituted_variables.append(f"# {key} = {val}")

    substituted_variables_str = "\n".join(substituted_variables)

    header = f"""\
################################################################################
# Variables introduced by Jhuki (version: {__version__})
################################################################################

{substituted_variables_str}

"""

    template_with_header = header + template

    with open(out_file, "w") as file_:
        file_.write(Template(template_with_header).substitute(sub_dict))


def write_many_parfiles_from_template(template, subs_dict, out_file_prefix):
    """Write several parfiles from the template with substitution specified by the
    dictionary sub_dict_list. The dictionary has to have as keys the variables to
    be substituted and as values:
    - "scalars" for those values that have to be the same in all the parfiles
    - lists, with one entry per file, for those that have to be change file by file.

    For example: {'dx': [1, 2], 'dy': 3} will create two files with
    'dx = 1' and 'dx = 2' and 'dy = 3' in both cases.

    We add an header with the list of variables that were introduced.

    :param template: Basic template with parameters that have to be
                     substituted. It follows Python's template
                     conventions, so the variables to be substituted
                     are to be prefixed with a dollar sign.
    :type template: str
    :param subs_dict: Dictionary that maps entries that have to be
                     substituted with their value, one per each file.
                     The values of the dictionary have to be lists or "scalars"
                     (if they have to have the same value for the files).
    :type subs_dict: dict with lists as values
    :param out_file_prefix: Path of the output files, without the extension ".par".
                            The parfiles will be saved as ``{out_file_prefix}1.par``,
                            ``{out_file_prefix}2.par``, and so on.
    :type out_file: str

    """

    # We check that all the lists with more than one element have the same length
    list_lengths = [
        len(v)
        for v in subs_dict.values()
        if (isinstance(v, list) and len(v) > 1)
    ]

    # Here we have greater than one because we allow for 0 elements
    if len(set(list_lengths)) > 1:
        raise ValueError("Substitution lists have different lengths.")

    if list_lengths:
        number_of_parfiles = list_lengths[0]
    else:
        number_of_parfiles = 1
    logging.debug(f"Writing {number_of_parfiles} parfiles")

    # Next, we ensure that all the elements are lists
    sub_dict_list = {
        k: (v if isinstance(v, list) else [v]) for k, v in subs_dict.items()
    }

    # Now we "broadcast" the scalar elements to list. With this,
    # sub_dict_broadcast will have lists with the same length
    # (= the number of parfiles)
    sub_dict_broadcast = {
        # else here is when the lists have only one element
        k: (v if len(v) == number_of_parfiles else v * number_of_parfiles)
        for k, v in sub_dict_list.items()
    }

    # Now we loop over the sub_dict_broadcast and create one parfile
    # for each entry
    for num_parfile in range(number_of_parfiles):
        file_name = f"{out_file_prefix}{num_parfile}.par"
        logging.debug(f"Writing {file_name}")
        subs = {k: v[num_parfile] for k, v in sub_dict_broadcast.items()}
        write_one_parfile_from_template(template, subs, file_name)



#!/usr/bin/env python3

# Copyright (C) 2020-2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.


"""The :py:mod:`~.nanchecker` module sets the parameter for NaNChecker.

This is a simple module and the only non-trivial function is
:py:func:`~.create_nanchecker_from_grid`, which automatically sets the
``check_every`` parameter.

"""

from functools import lru_cache

from jhuki.base import BaseThorn


class NaNChecker(BaseThorn):
    """Class that describes NaNChecker.

    This class is immutable.

    :ivar action_if_found: What to do if a NaN is found.
    :type action_if_found: str
    :ivar check_every: Check for NaNs every N iterations.
    :type check_every: int
    :ivar variables: List of variables for which NaNs have to be checked.
    :type variables: list of str

    """

    def __init__(self, variables, check_every, action_if_found="terminate"):
        """Constructor.

        :param check_every: Check for NaNs every N iterations.
        :type check_every: int
        :param variables: List of variables for which NaNs have to be checked.
        :type variables: list of str
        :param action_if_found: What to do if a NaN is found.
        :type action_if_found: str

        """
        self.variables = variables
        self.check_every = check_every
        self.action_if_found = action_if_found

    @property
    @lru_cache(1)
    def parfile_code(self):
        """Return the code you would put in your parfile."""

        vars_str = "\n".join(self.variables)

        return f"""\
NaNChecker::check_every = {self.check_every}
NaNChecker::action_if_found = {self.action_if_found}
NaNChecker::check_vars  = "{vars_str}"\
"""


def create_nanchecker_from_grid(grid, variables, action_if_found="terminate"):
    """Create a :py:class:`~.NaNChecker` object from a given :py:class:`~.Grid`.

    The only information used is when refinement levels are synced.

    :param grid: Simulation grid.
    :type grid: :py:class:`~.Grid`
    :param variables: List of variables for which NaNs have to be checked.
    :type variables: list of str
    :param action_if_found: What to do if a NaN is found.
    :type action_if_found: str

    :returns: NaNChecker object.
    :rtype: :py:class:`~.NaNChecker`

    """
    return NaNChecker(
        variables, grid.rl_synced_every, action_if_found=action_if_found
    )



#!/usr/bin/env python3

# Copyright (C) 2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.base` module contains basic functions and classes used
everywhere in the codebase.


"""

from abc import ABC, abstractmethod


class BaseThorn(ABC):
    """Base abstract class for thorn definitions.

    It defines the ``__str__`` method from the ``parfile_code`` one.

    Derived classes have to define a ``parfile_code`` method

    """

    @abstractmethod
    def parfile_code(self):  # pragma: no cover
        raise NotImplementedError

    def __str__(self):
        return self.parfile_code



# BSD 2-Clause License

# Copyright (c) 2018, Zachariah Etienne All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# As documented in the NRPyPN notebook
# PN-Hamiltonian-Spin-Spin.ipynb, this Python script
# generates spin-spin coupling pieces of the
# post-Newtonian (PN) Hamiltonian, up to and
# including 3PN order.

# Core functions:
# f_H_SS_2PN(m1,m2, S1U,S2U, nU, q):
#       Compute the complete H_SS_2PN term and store to
#                     global variable of the same name.
# f_HS1S2_3PN(m1,m2, n12U, S1U,S2U, p1U,p2U, q)):
#       Compute HS1S2_3PN and store to global variable
#                     of the same name.
# f_H_SS_S1sq_S2sq_3PN(m1,m2, n12U,n21U, S1U,S2U, p1U,p2U, q):
#       Compute H_SS_S1sq_S2sq_3PN and store to global
#                     variable of the same name.

# Author:  Zach Etienne
#          zachetie **at** gmail **dot* com

# Step 0: Add NRPy's directory to the path
# https://stackoverflow.com/questions/16780014/import-file-from-parent-directory
from . import (
    indexedexpNRPyPN as ixp,  # NRPy+: Symbolic indexed expression (e.g., tensors, vectors, etc.) support
)
from .NRPyPN_shortcuts import (  # NRPyPN: shortcuts for e.g., vector operations
    cross,
    div,
    dot,
)


#################################
#################################
# 2PN spin-spin term, from Eqs. 2.18 and 2.19 of
#      Buonanno, Chen, and Damour (2006):
#     https://arxiv.org/abs/gr-qc/0508067
def f_H_SS_2PN(m1, m2, S1U, S2U, nU, q):
    S0U = ixp.zerorank1()
    for i in range(3):
        S0U[i] = (1 + m2 / m1) * S1U[i] + (1 + m1 / m2) * S2U[i]
    global H_SS_2PN
    mu = m1 * m2 / (m1 + m2)
    H_SS_2PN = (
        mu / (m1 + m2) * (3 * dot(S0U, nU) ** 2 - dot(S0U, S0U)) / (2 * q ** 3)
    )


#################################
#################################
# 3PN spin-spin S_1,S_2 coupling term, from Eq. 2.11 of
#       Steinhoff, Hergt, and Sch\"afer (2008a)
#         https://arxiv.org/abs/0712.1716
def f_H_SS_S1S2_3PN(m1, m2, n12U, S1U, S2U, p1U, p2U, r12):
    global H_SS_S1S2_3PN
    H_SS_S1S2_3PN = (
        +div(3, 2) * (dot(cross(p1U, S1U), n12U) * dot(cross(p2U, S2U), n12U))
        + 6 * (dot(cross(p2U, S1U), n12U) * dot(cross(p1U, S2U), n12U))
        - 15
        * dot(S1U, n12U)
        * dot(S2U, n12U)
        * dot(p1U, n12U)
        * dot(p2U, n12U)
        - 3 * dot(S1U, n12U) * dot(S2U, n12U) * dot(p1U, p2U)
        + 3 * dot(S1U, p2U) * dot(S2U, n12U) * dot(p1U, n12U)
        + 3 * dot(S2U, p1U) * dot(S1U, n12U) * dot(p2U, n12U)
        + 3 * dot(S1U, p1U) * dot(S2U, n12U) * dot(p2U, n12U)
        + 3 * dot(S2U, p2U) * dot(S1U, n12U) * dot(p1U, n12U)
        - div(1, 2) * dot(S1U, p2U) * dot(S2U, p1U)
        + dot(S1U, p1U) * dot(S2U, p2U)
        - 3 * dot(S1U, S2U) * dot(p1U, n12U) * dot(p2U, n12U)
        + div(1, 2) * dot(S1U, S2U) * dot(p1U, p2U)
    ) / (2 * m1 * m2 * r12 ** 3)
    H_SS_S1S2_3PN += (
        (
            -dot(cross(p1U, S1U), n12U) * dot(cross(p1U, S2U), n12U)
            + dot(S1U, S2U) * dot(p1U, n12U) ** 2
            - dot(S1U, n12U) * dot(S2U, p1U) * dot(p1U, n12U)
        )
        * 3
        / (2 * m1 ** 2 * r12 ** 3)
    )
    H_SS_S1S2_3PN += (
        (
            -dot(cross(p2U, S2U), n12U) * dot(cross(p2U, S1U), n12U)
            + dot(S1U, S2U) * dot(p2U, n12U) ** 2
            - dot(S2U, n12U) * dot(S1U, p1U) * dot(p2U, n12U)
        )
        * 3
        / (2 * m2 ** 2 * r12 ** 3)
    )
    H_SS_S1S2_3PN += (
        (+dot(S1U, S2U) - 2 * dot(S1U, n12U) * dot(S2U, n12U))
        * 6
        * (m1 + m2)
        / r12 ** 4
    )


#################################
#################################
# 3PN spin-orbit coupling term, from Eq. 9 of
#    Steinhoff, Hergt, and Sch\"afer (2008b)
#       https://arxiv.org/abs/0809.2200
def f_H_SS_S1sq_S2sq_3PN(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r12):
    def f_H_SS_particle(
        m1, m2, n12U, S1U, _S2U, p1U, p2U, r12
    ):  # _S2U unused.
        H_SS_S1sq_S2sq_3PN_particle = (
            +m2 / (4 * m1 ** 3) * dot(p1U, S1U) ** 2
            + 3 * m2 / (8 * m1 ** 3) * dot(p1U, n12U) ** 2 * dot(S1U, S1U)
            - 3 * m2 / (8 * m1 ** 3) * dot(p1U, p1U) * dot(S1U, n12U) ** 2
            - 3
            * m2
            / (4 * m1 ** 3)
            * dot(p1U, n12U)
            * dot(S1U, n12U)
            * dot(p1U, S1U)
            - 3 / (4 * m1 * m2) * dot(p2U, p2U) * dot(S1U, S1U)
            + 9 / (4 * m1 * m2) * dot(p2U, p2U) * dot(S1U, n12U) ** 2
            + 3 / (4 * m1 ** 2) * dot(p1U, p2U) * dot(S1U, S1U)
            - 9 / (4 * m1 ** 2) * dot(p1U, p2U) * dot(S1U, n12U) ** 2
            - 3
            / (2 * m1 ** 2)
            * dot(p1U, n12U)
            * dot(p2U, S1U)
            * dot(S1U, n12U)
            + 3 / (m1 ** 2) * dot(p2U, n12U) * dot(p1U, S1U) * dot(S1U, n12U)
            + 3
            / (4 * m1 ** 2)
            * dot(p1U, n12U)
            * dot(p2U, n12U)
            * dot(S1U, S1U)
            - 15
            / (4 * m1 ** 2)
            * dot(p1U, n12U)
            * dot(p2U, n12U)
            * dot(S1U, n12U) ** 2
        ) / r12 ** 3
        H_SS_S1sq_S2sq_3PN_particle += (
            -(
                +div(9, 2) * dot(S1U, n12U) ** 2
                - div(5, 2) * dot(S1U, S1U)
                + 7 * m2 / m1 * dot(S1U, n12U) ** 2
                - 3 * m2 / m1 * dot(S1U, S1U)
            )
            * m2
            / r12 ** 4
        )
        return H_SS_S1sq_S2sq_3PN_particle

    global H_SS_S1sq_S2sq_3PN
    H_SS_S1sq_S2sq_3PN = +f_H_SS_particle(
        m1, m2, n12U, S1U, S2U, p1U, p2U, r12
    ) + f_H_SS_particle(m2, m1, n21U, S2U, S1U, p2U, p1U, r12)



# BSD 2-Clause License

# Copyright (c) 2018, Zachariah Etienne All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# As documented in the NRPyPN notebook
# PN-p_t.ipynb, this Python script
# generates the expression for the transverse
# component of momentum p_t up to and
# including terms at 3.5PN order.
# This is an implementation of the equations of
#  Ramos-Buades, Husa, and Pratten (2018)
#    https://arxiv.org/abs/1810.00036
# but validates against the relevant equation
#   in Healy, Lousto, Nakano, and Zlochower (2017)
#    https://arxiv.org/abs/1702.00872

# Core functions:
# f_p_t(m1,m2, chi1U,chi2U, r)
#       Compute p_t and store to
#       global variable of the same name.

# Author:  Zach Etienne
#          zachetie **at** gmail **dot* com

# Step 0: Add NRPy's directory to the path
# https://stackoverflow.com/questions/16780014/import-file-from-parent-directory
import sympy as sp  # SymPy: The Python computer algebra package upon which NRPy+ depends

from . import (
    indexedexpNRPyPN as ixp,  # NRPy+: Symbolic indexed expression (e.g., tensors, vectors, etc.) support
)
from .NRPyPN_shortcuts import (
    div,
)  # NRPyPN: shortcuts for e.g., vector operations

#################################
#################################
# Basic equation for p_t can be written in the
#   form:
# p_t = q/(sqrt(r)*(1+q)^2) (1 + \sum_{k=2}^7 (a_k/r^{k/2})
# where we construct the a_k terms in the sum below:

# Step 1: Construct terms a_2, a_3, and a_4, from
#  Eq A2 of Ramos-Buades, Husa, and Pratten (2018)
#    https://arxiv.org/abs/1810.00036
#  These terms have been independently validated
#    against the same terms in Eq 7 of
#  Healy, Lousto, Nakano, and Zlochower (2017)
#    https://arxiv.org/abs/1702.00872
def p_t__a_2_thru_a_4(m1, m2, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z):
    q = m2 / m1  # It is assumed that q >= 1, so m2 >= m1.
    global a_2, a_3, a_4
    a_2 = 2
    a_3 = -3 * (4 * q ** 2 + 3 * q) * chi2z / (4 * (q + 1) ** 2) - 3 * (
        3 * q + 4
    ) * chi1z / (4 * (q + 1) ** 2)
    a_4 = (
        -3 * q ** 2 * chi2x ** 2 / (2 * (q + 1) ** 2)
        + 3 * q ** 2 * chi2y ** 2 / (4 * (q + 1) ** 2)
        + 3 * q ** 2 * chi2z ** 2 / (4 * (q + 1) ** 2)
        + (+42 * q ** 2 + 41 * q + 42) / (16 * (q + 1) ** 2)
        - 3 * chi1x ** 2 / (2 * (q + 1) ** 2)
        - 3 * q * chi1x * chi2x / (q + 1) ** 2
        + 3 * chi1y ** 2 / (4 * (q + 1) ** 2)
        + 3 * q * chi1y * chi2y / (2 * (q + 1) ** 2)
        + 3 * chi1z ** 2 / (4 * (q + 1) ** 2)
        + 3 * q * chi1z * chi2z / (2 * (q + 1) ** 2)
    )


# Construct terms a_5 and a_6, from
#  Eq A2 of Ramos-Buades, Husa, and Pratten (2018)
#    https://arxiv.org/abs/1810.00036
#  These terms have been independently validated
#    against the same terms in Eq 7 of
#  Healy, Lousto, Nakano, and Zlochower (2017)
#    https://arxiv.org/abs/1702.00872
#  and a sign error was corrected in the a_5
#  expression.
def p_t__a_5_thru_a_6(
    m1, m2, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z, FixSignError=True
):
    SignFix = sp.sympify(-1)
    if FixSignError == False:
        SignFix = sp.sympify(+1)
    q = m2 / m1  # It is assumed that q >= 1, so m2 >= m1.
    global a_5, a_6
    a_5 = SignFix * (13 * q ** 3 + 60 * q ** 2 + 116 * q + 72) * chi1z / (
        16 * (q + 1) ** 4
    ) + (-72 * q ** 4 - 116 * q ** 3 - 60 * q ** 2 - 13 * q) * chi2z / (
        16 * (q + 1) ** 4
    )
    a_6 = (
        +(+472 * q ** 2 - 640) * chi1x ** 2 / (128 * (q + 1) ** 4)
        + (-512 * q ** 2 - 640 * q - 64) * chi1y ** 2 / (128 * (q + 1) ** 4)
        + (-108 * q ** 2 + 224 * q + 512) * chi1z ** 2 / (128 * (q + 1) ** 4)
        + (+472 * q ** 2 - 640 * q ** 4) * chi2x ** 2 / (128 * (q + 1) ** 4)
        + (+192 * q ** 3 + 560 * q ** 2 + 192 * q)
        * chi1x
        * chi2x
        / (128 * (q + 1) ** 4)
        + (-864 * q ** 3 - 1856 * q ** 2 - 864 * q)
        * chi1y
        * chi2y
        / (128 * (q + 1) ** 4)
        + (+480 * q ** 3 + 1064 * q ** 2 + 480 * q)
        * chi1z
        * chi2z
        / (128 * (q + 1) ** 4)
        + (-64 * q ** 4 - 640 * q ** 3 - 512 * q ** 2)
        * chi2y ** 2
        / (128 * (q + 1) ** 4)
        + (+512 * q ** 4 + 224 * q ** 3 - 108 * q ** 2)
        * chi2z ** 2
        / (128 * (q + 1) ** 4)
        + (
            +480 * q ** 4
            + 163 * sp.pi ** 2 * q ** 3
            - 2636 * q ** 3
            + 326 * sp.pi ** 2 * q ** 2
            - 6128 * q ** 2
            + 163 * sp.pi ** 2 * q
            - 2636 * q
            + 480
        )
        / (128 * (q + 1) ** 4)
    )


# Construct term a_7, from Eq A2 of
#  Ramos-Buades, Husa, and Pratten (2018)
#    https://arxiv.org/abs/1810.00036
def p_t__a_7(m1, m2, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z):
    q = m2 / m1  # It is assumed that q >= 1, so m2 >= m1.
    global a_7
    a_7 = (
        +5 * (4 * q + 1) * q ** 3 * chi2x ** 2 * chi2z / (2 * (q + 1) ** 4)
        - 5 * (4 * q + 1) * q ** 3 * chi2y ** 2 * chi2z / (8 * (q + 1) ** 4)
        - 5 * (4 * q + 1) * q ** 3 * chi2z ** 3 / (8 * (q + 1) ** 4)
        + chi1x
        * (
            +15 * (2 * q + 1) * q ** 2 * chi2x * chi2z / (4 * (q + 1) ** 4)
            + 15 * (1 * q + 2) * q * chi2x * chi1z / (4 * (q + 1) ** 4)
        )
        + chi1y
        * (
            +15 * q ** 2 * chi2y * chi1z / (4 * (q + 1) ** 4)
            + 15 * q ** 2 * chi2y * chi2z / (4 * (q + 1) ** 4)
        )
        + chi1z
        * (
            +15 * q ** 2 * (2 * q + 3) * chi2x ** 2 / (4 * (q + 1) ** 4)
            - 15 * q ** 2 * (q + 2) * chi2y ** 2 / (4 * (q + 1) ** 4)
            - 15 * q ** 2 * chi2z ** 2 / (4 * (q + 1) ** 3)
            - (
                103 * q ** 5
                + 145 * q ** 4
                - 27 * q ** 3
                + 252 * q ** 2
                + 670 * q
                + 348
            )
            / (32 * (q + 1) ** 6)
        )
        - (
            +348 * q ** 5
            + 670 * q ** 4
            + 252 * q ** 3
            - 27 * q ** 2
            + 145 * q
            + 103
        )
        * q
        * chi2z
        / (32 * (q + 1) ** 6)
        + chi1x ** 2
        * (
            +5 * (q + 4) * chi1z / (2 * (q + 1) ** 4)
            + 15 * q * (3 * q + 2) * chi2z / (4 * (q + 1) ** 4)
        )
        + chi1y ** 2
        * (
            -5 * (q + 4) * chi1z / (8 * (q + 1) ** 4)
            - 15 * q * (2 * q + 1) * chi2z / (4 * (q + 1) ** 4)
        )
        - 15 * q * chi1z ** 2 * chi2z / (4 * (q + 1) ** 3)
        - 5 * (q + 4) * chi1z ** 3 / (8 * (q + 1) ** 4)
    )


# Finally, sum the expressions for a_k to construct p_t as prescribed:
# p_t = q/(sqrt(r)*(1+q)^2) (1 + \sum_{k=2}^7 (a_k/r^{k/2}))
def f_p_t(m1, m2, chi1U, chi2U, r):
    q = m2 / m1  # It is assumed that q >= 1, so m2 >= m1.
    a = ixp.zerorank1(DIM=10)
    p_t__a_2_thru_a_4(
        m1, m2, chi1U[0], chi1U[1], chi1U[2], chi2U[0], chi2U[1], chi2U[2]
    )
    a[2] = a_2
    a[3] = a_3
    a[4] = a_4
    p_t__a_5_thru_a_6(
        m1, m2, chi1U[0], chi1U[1], chi1U[2], chi2U[0], chi2U[1], chi2U[2]
    )
    a[5] = a_5
    a[6] = a_6
    p_t__a_7(
        m1, m2, chi1U[0], chi1U[1], chi1U[2], chi2U[0], chi2U[1], chi2U[2]
    )
    a[7] = a_7
    global p_t
    p_t = 1  # Term prior to the sum in parentheses
    for k in range(8):
        p_t += a[k] / r ** div(k, 2)
    p_t *= q / (1 + q) ** 2 * 1 / r ** div(1, 2)



#!/usr/bin/env python3

# Copyright (C) 2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.


"""The :py:mod:`~.horizons` module provides infrastructure to work with horizons.

"""
from jhuki.base import BaseThorn

class Horizon(BaseThorn):
    """
    """
    def __init__(ah_num, spherical_surface_num):
        self.spherical_surface_num =
        pass

    @property
    @lru_cache(1)
    def parfile_code():
        """Return the code you would put in your parfile."""

        def assign_parameter(param, value):
            return (
                f"AHFinderDirect::{param}[{self.ah_num}] = {value}"
            )

        ret = []



#!/usr/bin/env python3

# Copyright (C) 2020-2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.


"""The :py:mod:`~.dissipation` module takes care of setting parameters for the
dissipation thorn. The main class is :py:class:`~.Dissipation`, but it is more
convenient to use helper functions to setup this object. For example, to set
up dissipation for a simulation with Lean, the simplest way is:

.. code-block:: python

   dis = add_gauge(add_Lean(create_dissipation_from_grid(grid, 0.3)))

where 0.3 is the value of ``eps_dis`` at the finest level.

"""

import enum
from functools import lru_cache

from jhuki.base import BaseThorn


class Dissipation(BaseThorn):
    """Class that describes the dissipation. Best used in combination with helper
    functions.

    This class is immutable.

    :ivar epsdis_per_level: Dictionary that maps the value of the dissipation on
                            each refinement level.
    :type epsdis_per_level: dict
    :ivar order: Order of the dissipation.
    :type order: int
    :ivar variables: List of variables to which the dissipation has to be applied.
    :type variables: list of str

    """

    def __init__(self, epsdis_per_level, order, variables):
        """It is your responsibility to provide meaningful eps_dis_per_level.

        :param epsdis_per_level: Dictionary that maps the value of the
                                 dissipation on each refinement level.
        :type epsdis_per_level: dict
        :param order: Order of the dissipation.
        :type order: int
        :param variables: List of variables to which the dissipation has to be
                          applied.
        :type variables: list of str

        """
        self.epsdis_per_level = epsdis_per_level
        self.order = order
        self.variables = variables

    @property
    @lru_cache(1)
    def parfile_code(self):
        """Return the code you would put in your parfile."""

        eps_dis_str = "\n".join(
            f"Dissipation::epsdis_for_level[{level}] = {epsdis}"
            for level, epsdis in self.epsdis_per_level.items()
        )

        vars_str = "\n".join(self.variables)

        return f"""\
Dissipation::order = {self.order}
Dissipation::vars  = "{vars_str}"
{eps_dis_str}"""


class DissPrescription(enum.Enum):
    """DissPrescription for how to set up the dissipation.

    Three different prescriptions are available:
    - ``const``: set the same ``eps_dis`` everywhere,
    - ``dtfac``: set ``eps_dis`` proportional to the local dtfac,
    - ``continuous``: set ``eps_dis`` proportional to the local grid spacing
                      to the power of the order

    """

    const = enum.auto()
    dtfact = enum.auto()
    continuous = enum.auto()


def create_dissipation_from_grid(
    grid, eps_dis_finest, variables=None, prescription=DissPrescription.dtfact
):
    """Create a :py:class:`~.Dissipation` object from a given :py:class:`~.Grid`.


    The order of the dissipation is deduced from the number of ghost zones.

    :param grid: Simulation grid.
    :type grid: :py:class:`~.Grid`
    :param epsdis_per_level: Dictionary that maps the value of the dissipation
                             on each refinement level.
    :type epsdis_per_level: dict
    :param order: Order of the dissipation.
    :type order: int
    :param variables: List of variables to which the dissipation has to be
                      applied.
    :type variables: list of str
    :param prescription: How to set ``eps_dis``.
    :type prescription: :py:class:`~.DissPrescription`

    :returns: Dissipation object.
    :rtype: :py:class:`~.Dissipation`

    """
    if not isinstance(prescription, DissPrescription):
        raise ValueError(
            f"Unknown prescription {prescription}. "
            "Accepted values: {list(DissPrescription.__members__.keys())}"
        )

    order = 2 * grid.num_ghost - 1
    variables = [] if variables is None else variables

    # If we have N levels, we need to set up dissipation on N + 1 zones
    num_levels = grid.max_num_refinement_levels + 1

    if prescription == DissPrescription.const:
        # Same everywhere
        eps_dis_levels = {level: eps_dis_finest for level in range(num_levels)}
    elif prescription == DissPrescription.dtfact:
        # Same everywhere, halved by two every time a level is synced
        # For example, if have 4 levels, 2 synced, then we should have
        # [1/4, 1/2, 1, 1] * eps_dis

        num_levels_synced = grid.num_levels_with_dt_coarse

        # First, the ones with the values halved
        eps_dis_levels = {
            level: eps_dis_finest / 2 ** (num_levels_synced - level)
            for level in range(num_levels_synced)
        }

        # Next, we add all the ones that are eps_dis_finest
        eps_dis_levels.update(
            {
                level: eps_dis_finest
                for level in range(
                    num_levels_synced,
                    num_levels,
                )
            }
        )

    elif prescription == DissPrescription.continuous:
        # Proportional to the local size, so there's a factor of 2**order for every
        # refinement level
        eps_dis_levels = {
            level: eps_dis_finest / (2 ** (num_levels - level - 1)) ** order
            for level in range(num_levels)
        }

    return Dissipation(eps_dis_levels, order, variables)


def add_to_dissipation(variables, doc):
    """Return a function that adds variables to the given Dissipation."""

    def inner(dissipation):
        __doc__ = doc  # noqa: F841
        return Dissipation(
            dissipation.epsdis_per_level,
            dissipation.order,
            dissipation.variables + variables,
        )

    return inner


_lean_vars = [
    "LeanBSSNMoL::conf_fac",
    "LeanBSSNMoL::hmetric",
    "LeanBSSNMoL::hcurv",
    "LeanBSSNMoL::trk",
    "LeanBSSNMoL::gammat",
]

_gauge_vars = ["ADMBase::lapse", "ADMBase::shift"]

_proca_vars = [
    "ProcaBase::Ei",
    "ProcaBase::Ai",
    "ProcaBase::Aphi",
    "ProcaBase::Zeta",
]

add_Lean = add_to_dissipation(
    _lean_vars,
    """Return a new :py:class:`~.Dissipation` with Lean's variables added.

    :param dissipation: Dissipation that has to be used as base.
    :type dissipation: :py:class:`~.Dissipation`
    :returns: Dissipation with Lean's variables added.
    :rtype: :py:class:`~.Dissipation`""",
)

add_gauge = add_to_dissipation(
    _gauge_vars,
    """Return a new :py:class:`~.Dissipation` with guage variables added.

    :param dissipation: Dissipation that has to be used as base.
    :type dissipation: :py:class:`~.Dissipation`
    :returns: Dissipation with guage variables added.
    :rtype: :py:class:`~.Dissipation`""",
)

add_Proca = add_to_dissipation(
    _proca_vars,
    """Return a new :py:class:`~.Dissipation` with Proca variables added.

    :param dissipation: Dissipation that has to be used as base.
    :type dissipation: :py:class:`~.Dissipation`
    :returns: Dissipation with Proca variables added.
    :rtype: :py:class:`~.Dissipation`""",
)



#!/usr/bin/env python3

# Copyright (C) 2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.twochargedpunctures` module extends :py:mod:`~.twopunctures`
 for the case of charged black holes. The interface is nearly identical.
"""

from functools import lru_cache
from math import sqrt

from jhuki.externals.nrpypn import compute_quasicircular_momenta
from jhuki.twopunctures import TwoPunctures


def prepare_quasicircular_inspiral(
    mass_ratio,
    coordinate_distance,
    total_bare_mass=1,
    chi_plus=(0, 0, 0),
    chi_minus=(0, 0, 0),
    lambda_plus=0,
    lambda_minus=0,
    **kwargs,
):
    """Return a :py:class:`~.TwoChargedPunctures` that describes a quasi-circular inspiral.
    We always assume that the plus puncture is the most massive one.

    We compute the momenta starting from the uncharged case and scaling by
    sqrt(1 - lambda1 * lambda2). This is essentially 0 PN. It works well for low
    charge, but not that much for higher charge.

    :param mass_ratio: Mass ratio of the binary.
    :type mass_ratio: float
    :param lambda_plus: Charge-to-mass-ratio of the black hole on the positive
                        side of the x (or z) axis.
    :type lambda_plus: float
    :param lambda_minus: Charge-to-mass-ratio of the black hole on the negative
                         side of the x (or z) axis.
    :type lambda_minus: float
    :param coordinate_distance: Initial coordinate separation.
    :type coordinate_distance: float
    :param total_bare_mass: Rescale masses and momenta so that the total bare mass of the
                            system is this.
    :type total_bare_mass: float

    :param chi_plus: Dimensionless spin of the black hole on the positive side of the x
                     (or z) axis along the three directions.
    :type chi_plus: tuple/list with three numbers

    :param chi_minus: Dimensionless spin of the black hole on the negative side of the x
                     (or z) axis along the three directions.
    :type chi_minus: tuple/list with three numbers

    Unknown arguments are passed to :py:class:`~.TwoPunctures`.

    :returns: A :py:class:`~.TwoPunctures` for a quasi-circular inspiral.
    :rtype: :py:class:`~.TwoPunctures`

    """

    if mass_ratio < 1:
        mass_ratio = 1 / mass_ratio

    Pt, Pr = compute_quasicircular_momenta(
        mass_ratio, coordinate_distance, chi_plus, chi_minus
    )

    mass_plus = mass_ratio / (1 + mass_ratio) * total_bare_mass
    mass_minus = 1 / (1 + mass_ratio) * total_bare_mass

    factor = sqrt(1 - lambda_plus * lambda_minus)

    Pt, Pr = Pt * total_bare_mass * factor, Pr * total_bare_mass * factor

    momenta_plus = (-Pr, Pt, 0)
    momenta_minus = (Pr, -Pt, 0)

    return TwoChargedPunctures(
        mass_plus,
        mass_minus,
        coordinate_distance,
        momenta_plus=momenta_plus,
        momenta_minus=momenta_minus,
        chi_plus=chi_plus,
        chi_minus=chi_minus,
        charge_plus=lambda_plus * mass_plus,
        charge_minus=lambda_minus * mass_minus,
        give_bare_mass=False,
        **kwargs,
    )


class TwoChargedPunctures(TwoPunctures):
    """The :py:class:`~.TwoChargedPunctures` extends :py:class:`~.TwoPunctures` for
    the case charged black holes (to use with the ``TwoChargedPunctures`` thorn).

    As :py:class:`~.TwoPunctures`, this class is immutable.

    :ivar charge_plus: Charge of the black hole on the positive side of the x (or z) axis.
    :vartype charge_plus: float
    :ivar charge_minus: Charge of the black hole on the negative side of the x (or z) axis.
    :vartype charge_minus: float
    """

    def __init__(
        self,
        mass_plus,
        mass_minus,
        coordinate_distance,
        momenta_plus=None,
        momenta_minus=None,
        chi_plus=None,
        chi_minus=None,
        charge_plus=0,
        charge_minus=0,
        swap_xz=False,
        give_bare_mass=False,
    ):
        """Constructor.

        :param mass_plus: Mass of the black hole on the positive side of the x (or z) axis.
        :type mass_plus: float
        :param mass_minus: Mass of the black hole on the negative side of the x (or z) axis.
        :type mass_minus: float

        :param charge_plus: Charge of the black hole on the positive side of the x (or z) axis.
        :type charge_plus: float
        :param charge_minus: Charge of the black hole on the negative side of the x (or z) axis.
        :type charge_minus: float

        :param coordinate_separation: Initial coordinate separation.
        :type coordinate_separation: float
        :param par_b: Half of the initial separation.
        :type par_b: float

        :param momenta_plus: Array with the linear momenta along the three directions for the
                            black hole on the positive side of the x (or z) axis.
                             If ``swap_xz`` is True, these are the values after swapping.
                             For example ``(0, 0, 0.5)`` is a vector along the z axis.
                             (This is the opposite of what happens in ``TwoPunctures``, where
                              you have to define the values before swapping)
        :type momenta_plus: list/tuple
        :param momenta_minus: Array with the linear momenta along the three directions for the
                             black hole on the positive side of the x (or z) axis.
                             If ``swap_xz`` is True, these are the values after swapping.
                             For example ``(0, 0, 0.5)`` is a vector along the z axis.
                             (This is the opposite of what happens in ``TwoPunctures``, where
                              you have to define the values before swapping)
        :type momenta_minus: list/tuple

        :param chi_plus: Array with the dimensionless_spin along the three directions for the
                        black hole on the positive side of the x (or z) axis.
                             If ``swap_xz`` is True, these are the values after swapping.
                             For example ``(0, 0, 0.5)`` is a vector along the z axis.
                             (This is the opposite of what happens in ``TwoPunctures``, where
                              you have to define the values before swapping)
        :type chi_plus: list/tuple
        :param chi_minus: Array with the dimensionless spin along the three directions for the
                         black hole on the positive side of the x (or z) axis.
                             If ``swap_xz`` is True, these are the values after swapping.
                             For example ``(0, 0, 0.5)`` is a vector along the z axis.
                             (This is the opposite of what happens in ``TwoPunctures``, where
                              you have to define the values before swapping)
        :type chi_minus: list/tuple

        :param swap_xz: If True, activate the ``swap_xz`` parameter in TwoPunctures.
        :type swap_xz: bool

        :param give_bare_mass: If True, set this parameter to True in the parfile.
        :type give_bare_mass: bool
        """

        super().__init__(
            mass_plus,
            mass_minus,
            coordinate_distance,
            momenta_plus=momenta_plus,
            momenta_minus=momenta_minus,
            chi_plus=chi_plus,
            chi_minus=chi_minus,
            swap_xz=swap_xz,
            give_bare_mass=give_bare_mass,
            initial_alpha="psi^n",
        )

        self.charge_plus = charge_plus
        self.charge_minus = charge_minus

    @property
    @lru_cache(1)
    def parfile_code(self):
        """Return the code you would put in your parfile."""

        uncharged_parfile = (
            super()
            .parfile_code.replace("TwoPunctures", "TwoChargedPunctures")
            .replace("twopunctures", "twochargedpunctures")
        )

        charged_parfile = (
            f"{uncharged_parfile}\n"
            f"TwoChargedPunctures::par_q_plus = {self.charge_plus}\n"
            f"TwoChargedPunctures::par_q_minus = {self.charge_minus}"
        )

        return charged_parfile



# BSD 2-Clause License

# Copyright (c) 2018, Zachariah Etienne All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# As documented in the NRPyPN notebook
# PN-Hamiltonian-Spin-Orbit.ipynb, this Python script
# generates spin-orbit coupling pieces of the
# post-Newtonian (PN) Hamiltonian, up to and
# including 3.5PN order.

# Core functions:
# f_H_SO_2p5PN(m1,m2, n12U,n21U, S1U, S2U, p1U,p2U, r12):
#       Compute H_SO_1p5PN and store to global variable
#                     of the same name.
# f_H_SO_2p5PN(m1,m2, n12U,n21U, S1U, S2U, p1U,p2U, r12):
#       Compute H_SO_2p5PN and store to global variable
#                     of the same name.
# f_H_SO_3p5PN(m1,m2, n12U,n21U, S1U, S2U, p1U,p2U, r12):
#       Compute H_SO_3p5PN and store to global variable
#                     of the same name.

# Author:  Zach Etienne
#          zachetie **at** gmail **dot* com

# Step 0: Add NRPy's directory to the path
# https://stackoverflow.com/questions/16780014/import-file-from-parent-directory
from . import (
    indexedexpNRPyPN as ixp,  # NRPy+: Symbolic indexed expression (e.g., tensors, vectors, etc.) support
)
from .NRPyPN_shortcuts import (  # NRPyPN: shortcuts for e.g., vector operations
    cross,
    div,
    dot,
)

#################################
#################################

# 1.5PN spin-orbit coupling term, from Eq. 4.11a of
#    Damour, Jaranowski, and Sch\"afer (2008)
#      https://arxiv.org/abs/0711.1048
def f_H_SO_1p5PN(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r12):
    def f_Omega1(m1, m2, n12U, p1U, p2U, r12):
        Omega1 = ixp.zerorank1()
        for i in range(3):
            Omega1[i] = (
                div(3, 2) * m2 / m1 * cross(n12U, p1U)[i]
                - 2 * cross(n12U, p2U)[i]
            ) / r12 ** 2
        return Omega1

    global H_SO_1p5PN
    Omega1 = f_Omega1(m1, m2, n12U, p1U, p2U, r12)
    Omega2 = f_Omega1(m2, m1, n21U, p2U, p1U, r12)
    H_SO_1p5PN = dot(Omega1, S1U) + dot(Omega2, S2U)


#################################
#################################

# 2.5PN spin-orbit coupling term, from Eq. 4.11b of
#    Damour, Jaranowski, and Sch\"afer (2008)
#      https://arxiv.org/abs/0711.1048
def f_H_SO_2p5PN(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r12):
    def f_Omega_SO_2p5PN(m1, m2, n12U, p1U, p2U, r12):
        Omega1 = ixp.zerorank1()
        for i in range(3):
            Omega1[i] = (
                +(
                    +(-div(11, 2) * m2 - 5 * m2 ** 2 / m1)
                    * cross(n12U, p1U)[i]
                    + (6 * m1 + div(15, 2) * m2) * cross(n12U, p2U)[i]
                )
                / r12 ** 3
                + (
                    +(
                        -div(5, 8) * m2 * dot(p1U, p1U) / m1 ** 3
                        - div(3, 4) * dot(p1U, p2U) / m1 ** 2
                        + div(3, 4) * dot(p2U, p2U) / (m1 * m2)
                        - div(3, 4) * dot(n12U, p1U) * dot(n12U, p2U) / m1 ** 2
                        - div(3, 2) * dot(n12U, p2U) ** 2 / (m1 * m2)
                    )
                    * cross(n12U, p1U)[i]
                    + (
                        dot(p1U, p2U) / (m1 * m2)
                        + 3 * dot(n12U, p1U) * dot(n12U, p2U) / (m1 * m2)
                    )
                    * cross(n12U, p2U)[i]
                    + (
                        div(3, 4) * dot(n12U, p1U) / m1 ** 2
                        - 2 * dot(n12U, p2U) / (m1 * m2)
                    )
                    * cross(p1U, p2U)[i]
                )
                / r12 ** 2
            )
        return Omega1

    Omega1_2p5PNU = f_Omega_SO_2p5PN(m1, m2, n12U, p1U, p2U, r12)
    Omega2_2p5PNU = f_Omega_SO_2p5PN(m2, m1, n21U, p2U, p1U, r12)

    global H_SO_2p5PN
    H_SO_2p5PN = dot(Omega1_2p5PNU, S1U) + dot(Omega2_2p5PNU, S2U)


#################################
#################################
# 3.5PN spin-orbit coupling term, from Eq. 5 of
#    Hartung and Steinhoff (2011)
#   https://arxiv.org/abs/1104.3079

# 3.5PN H_SO:  Omega_1, part 1:
def HS2011_Omega_SO_3p5PN_pt1(m1, m2, n12U, p1U, p2U, r12):
    Omega1 = ixp.zerorank1()
    for i in range(3):
        Omega1[i] = (
            (
                +7 * m2 * dot(p1U, p1U) ** 2 / (16 * m1 ** 5)
                + 9
                * dot(n12U, p1U)
                * dot(n12U, p2U)
                * dot(p1U, p1U)
                / (16 * m1 ** 4)
                + 3 * dot(p1U, p1U) * dot(n12U, p2U) ** 2 / (4 * m1 ** 3 * m2)
                + 45
                * dot(n12U, p1U)
                * dot(n12U, p2U) ** 3
                / (16 * m1 ** 2 * m2 ** 2)
                + 9 * dot(p1U, p1U) * dot(p1U, p2U) / (16 * m1 ** 4)
                - 3
                * dot(n12U, p2U) ** 2
                * dot(p1U, p2U)
                / (16 * m1 ** 2 * m2 ** 2)
                - 3 * dot(p1U, p1U) * dot(p2U, p2U) / (16 * m1 ** 3 * m2)
                - 15
                * dot(n12U, p1U)
                * dot(n12U, p2U)
                * dot(p2U, p2U)
                / (16 * m1 ** 2 * m2 ** 2)
                + 3 * dot(n12U, p2U) ** 2 * dot(p2U, p2U) / (4 * m1 * m2 ** 3)
                - 3 * dot(p1U, p2U) * dot(p2U, p2U) / (16 * m1 ** 2 * m2 ** 2)
                - 3 * dot(p2U, p2U) ** 2 / (16 * m1 * m2 ** 3)
            )
            * cross(n12U, p1U)[i]
        ) / r12 ** 2
    return Omega1


# 3.5PN H_SO:  Omega_1, part 2:
def HS2011_Omega_SO_3p5PN_pt2(m1, m2, n12U, p1U, p2U, r12):
    Omega1 = ixp.zerorank1()
    for i in range(3):
        Omega1[i] = (
            +(
                -3
                * dot(n12U, p1U)
                * dot(n12U, p2U)
                * dot(p1U, p1U)
                / (2 * m1 ** 3 * m2)
                - 15
                * dot(n12U, p1U) ** 2
                * dot(n12U, p2U) ** 2
                / (4 * m1 ** 2 * m2 ** 2)
                + 3
                * dot(p1U, p1U)
                * dot(n12U, p2U) ** 2
                / (4 * m1 ** 2 * m2 ** 2)
                - dot(p1U, p1U) * dot(p1U, p2U) / (2 * m1 ** 3 * m2)
                + dot(p1U, p2U) ** 2 / (2 * m1 ** 2 * m2 ** 2)
                + 3
                * dot(n12U, p1U) ** 2
                * dot(p2U, p2U)
                / (4 * m1 ** 2 * m2 ** 2)
                - dot(p1U, p1U) * dot(p2U, p2U) / (4 * m1 ** 2 * m2 ** 2)
                - 3
                * dot(n12U, p1U)
                * dot(n12U, p2U)
                * dot(p2U, p2U)
                / (2 * m1 * m2 ** 3)
                - dot(p1U, p2U) * dot(p2U, p2U) / (2 * m1 * m2 ** 3)
            )
            * cross(n12U, p2U)[i]
        ) / r12 ** 2
    return Omega1


# 3.5PN H_SO:  Omega_1, part 3:
def HS2011_Omega_SO_3p5PN_pt3(m1, m2, n12U, p1U, p2U, r12):
    Omega1 = ixp.zerorank1()
    for i in range(3):
        Omega1[i] = (
            +(
                -9 * dot(n12U, p1U) * dot(p1U, p1U) / (16 * m1 ** 4)
                + dot(p1U, p1U) * dot(n12U, p2U) / (m1 ** 3 * m2)
                + 27
                * dot(n12U, p1U)
                * dot(n12U, p2U) ** 2
                / (16 * m1 ** 2 * m2 ** 2)
                - dot(n12U, p2U) * dot(p1U, p2U) / (8 * m1 ** 2 * m2 ** 2)
                - 5 * dot(n12U, p1U) * dot(p2U, p2U) / (16 * m1 ** 2 * m2 ** 2)
            )
            * cross(p1U, p2U)[i]
        ) / r12 ** 2
    return Omega1


# 3.5PN H_SO:  Omega_1, part 4:
def HS2011_Omega_SO_3p5PN_pt4(m1, m2, n12U, p1U, p2U, r12):
    Omega1 = ixp.zerorank1()
    for i in range(3):
        Omega1[i] = (
            +(
                -3 * m2 * dot(n12U, p1U) ** 2 / (2 * m1 ** 2)
                + ((-3 * m2) / (2 * m1 ** 2) + 27 * m2 ** 2 / (8 * m1 ** 3))
                * dot(p1U, p1U)
                + (177 / (16 * m1) + 11 / m2) * dot(n12U, p2U) ** 2
                + (11 / (2 * m1) + 9 * m2 / (2 * m1 ** 2))
                * dot(n12U, p1U)
                * dot(n12U, p2U)
                + (23 / (4 * m1) + 9 * m2 / (2 * m1 ** 2)) * dot(p1U, p2U)
                - (159 / (16 * m1) + 37 / (8 * m2)) * dot(p2U, p2U)
            )
            * cross(n12U, p1U)[i]
        ) / r12 ** 3
    return Omega1


# 3.5PN H_SO:  Omega_1, part 5:
def HS2011_Omega_SO_3p5PN_pt5(m1, m2, n12U, p1U, p2U, r12):
    Omega1 = ixp.zerorank1()
    for i in range(3):
        Omega1[i] = (
            +(
                +4 * dot(n12U, p1U) ** 2 / m1
                + 13 * dot(p1U, p1U) / (2 * m1)
                + 5 * dot(n12U, p2U) ** 2 / m2
                + 53 * dot(p2U, p2U) / (8 * m2)
                - (211 / (8 * m1) + 22 / m2) * dot(n12U, p1U) * dot(n12U, p2U)
                - (47 / (8 * m1) + 5 / m2) * dot(p1U, p2U)
            )
            * cross(n12U, p2U)[i]
        ) / r12 ** 3
    return Omega1


# 3.5PN H_SO:  Omega_1, part 6:
def HS2011_Omega_SO_3p5PN_pt6(m1, m2, n12U, p1U, p2U, r12):
    Omega1 = ixp.zerorank1()
    for i in range(3):
        Omega1[i] = (
            +(
                -(8 / m1 + 9 * m2 / (2 * m1 ** 2)) * dot(n12U, p1U)
                + (59 / (4 * m1) + 27 / (2 * m2)) * dot(n12U, p2U)
            )
            * cross(p1U, p2U)[i]
        ) / r12 ** 3
    return Omega1


# 3.5PN H_SO:  Omega_1, part 7:
def HS2011_Omega_SO_3p5PN_pt7(m1, m2, n12U, p1U, p2U, r12):
    Omega1 = ixp.zerorank1()
    for i in range(3):
        Omega1[i] = (
            +(181 * m1 * m2 / 16 + 95 * m2 ** 2 / 4 + 75 * m2 ** 3 / (8 * m1))
            * cross(n12U, p1U)[i]
            - (21 * m1 ** 2 / 2 + 473 * m1 * m2 / 16 + 63 * m2 ** 2 / 4)
            * cross(n12U, p2U)[i]
        ) / r12 ** 4
    return Omega1


# 3.5PN H_SO: Combining all the above Omega terms
#             into the full 3.5PN S-O Hamiltonian
#             expression
def f_H_SO_3p5PN(m1, m2, n12U, n21U, S1U, S2U, p1U, p2U, r12):
    Omega1_3p5PNU = ixp.zerorank1()
    Omega2_3p5PNU = ixp.zerorank1()
    for i in range(3):
        Omega1_3p5PNU[i] = HS2011_Omega_SO_3p5PN_pt1(
            m1, m2, n12U, p1U, p2U, r12
        )[i]
        Omega1_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt2(
            m1, m2, n12U, p1U, p2U, r12
        )[i]
        Omega1_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt3(
            m1, m2, n12U, p1U, p2U, r12
        )[i]
        Omega1_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt4(
            m1, m2, n12U, p1U, p2U, r12
        )[i]
        Omega1_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt5(
            m1, m2, n12U, p1U, p2U, r12
        )[i]
        Omega1_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt6(
            m1, m2, n12U, p1U, p2U, r12
        )[i]
        Omega1_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt7(
            m1, m2, n12U, p1U, p2U, r12
        )[i]

        Omega2_3p5PNU[i] = HS2011_Omega_SO_3p5PN_pt1(
            m2, m1, n21U, p2U, p1U, r12
        )[i]
        Omega2_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt2(
            m2, m1, n21U, p2U, p1U, r12
        )[i]
        Omega2_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt3(
            m2, m1, n21U, p2U, p1U, r12
        )[i]
        Omega2_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt4(
            m2, m1, n21U, p2U, p1U, r12
        )[i]
        Omega2_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt5(
            m2, m1, n21U, p2U, p1U, r12
        )[i]
        Omega2_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt6(
            m2, m1, n21U, p2U, p1U, r12
        )[i]
        Omega2_3p5PNU[i] += HS2011_Omega_SO_3p5PN_pt7(
            m2, m1, n21U, p2U, p1U, r12
        )[i]

    global H_SO_3p5PN
    H_SO_3p5PN = dot(Omega1_3p5PNU, S1U) + dot(Omega2_3p5PNU, S2U)



#!/usr/bin/env python3

# Copyright (C) 2020 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""Jhuki provides utilities to prepare simulations for the Einstein Toolkit.

"""

__version__ = "0.1.0-dev9"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jhuki', 'jhuki.externals', 'jhuki.externals.nrpypn']

package_data = \
{'': ['*']}

install_requires = \
['ConfigArgParse>=1.2.3,<2.0.0', 'sympy>=1.7.1,<2.0.0']

setup_kwargs = {
    'name': 'jhuki',
    'version': '0.1.0.dev9',
    'description': 'Utilities to prepare Einstein Toolkit simulations',
    'long_description': '<p align="center">\n<img src="https://github.com/Sbozzolo/jhuki/raw/master/logo.png" height="250">\n</p>\n\n[![GPLv3\nlicense](https://img.shields.io/badge/License-GPLv3-blue.svg)](http://perso.crans.org/besson/LICENSE.html)\n![Tests](https://github.com/Sbozzolo/Jhuki/workflows/Tests/badge.svg)\n[![PyPI version](https://badge.fury.io/py/jhuki.svg)](https://badge.fury.io/py/Jhuki)\n[![codecov](https://codecov.io/gh/Sbozzolo/Jhuki/branch/master/graph/badge.svg)](https://codecov.io/gh/Sbozzolo/Jhuki)\n[![Total alerts](https://img.shields.io/lgtm/alerts/g/Sbozzolo/Jhuki.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/Sbozzolo/Jhuki/alerts/)\n\nWriting parameter files for Cactus simulations is not easy. To achieve a\nsuccessful evolution, one has to tune several parameters, which typically depend\non the grid configuration or on other settings. Writing these par files by hand\nis tedious and error-prone. `Jhuki` is a Python library to prepare\nsimulations for the `Einstein Toolkit` (or Cactus-based codes).\n\n> :warning: This package is currently under development. It may be full of bugs,\n>           and its interfaces might change without notice. It is also highly\n>           opinionated.\n\n## Features\n\n* Generate parameter files from a template using configuration files or\n  command-line arguments\n* Take care of the grid configuration given the desired resolution at the finest\n  level and other details\n* Generate binary black hole configurations for quasi-circular mergers\n  (automatically setting the linear momenta)\n* Add dissipation\n\nYou can find the reference material at the website\n[sbozzolo.github.io/Jhuki](https://sbozzolo.github.io/Jhuki).\n\n## Examples\n\n\n### Integration with SimFactory\n\n`Jhuki` integrates well with `SimFactory`. Assume you have the following\nsimple `test_sim.rpar` file:\n``` python\n#!/usr/bin/env python3\n\nfrom jhuki.simfactory import simfactory_option\nfrom jhuki.simfactory import write_parfile\n\npar_b = simfactory_option("@PARB@", 5)\n\nm_plus = 0.5\n\nlines="""\nTwoPunctures::par_b      = $par_b\nTwoPunctures::par_m_plus = $m_plus\n\n# Rest of the parameter file\n\n"""\n\nwrite_parfile(lines)\n```\n\n`SimFactory` can execute this Python file to produce a `.par` file to run:\n\n``` sh\nsim create-run test_sim.rpar\n```\nwill run the parfile with content\n``` sh\nTwoPunctures::par_b      = 5\nTwoPunctures::par_m_plus = 0.5\n```\nThe value for `par_b` is run-time adjustable:\n``` sh\nsim create-run test_sim.rpar --define PARB 10\n```\nwill produce\n``` sh\nTwoPunctures::par_b      = 10\nTwoPunctures::par_m_plus = 0.5\n```\nSince arbitrary Python code can be executed before the parfile is written,\nall the features of `Jhuki` presented below can be used directly. For instance,\nas discussed below, one can prepare quasi circular inspirals with the function\n`prepare_quasicircular_inspiral`, so one can add the entire parfile code for\nthe two puncture with:\n``` python\n#!/usr/bin/env python3\n\nfrom jhuki.simfactory import simfactory_option\nfrom jhuki.simfactory import write_parfile\nfrom jhuki.twopunctures import prepare_quasicircular_inspiral\n\npar_b = simfactory_option("@PARB@", 5)\n\ntwopunctures = prepare_quasicircular_inspiral(mass_ratio=1,\n                                              coordinate_distance=par_b)\n\nlines="""\n\n$twopunctures\n\n# Rest of the parameter file\n"""\n\nwrite_parfile(lines)\n\n```\nThis sets up a good fraction of the parfile needed to run quasi circular\ninspirals.\n\n### Working with grid structures\n\nProblem: you want to generate the parfile code for a grid structure with two\nrefinement centers, each with 8 levels and resolution at the finest level 0.001\nand CFL factor of 0.4 (in the finest level). In this, you want to ensure that\nthe maximum timestep on the grid never exceeds 1 to avoid numerical instability.\n\n``` python\n#!/usr/bin/env python3\n\nfrom jhuki import grid as pg\n\nrefinement_radii = tuple(2**level for level in range(7))\n\ncenter1 = pg.RefinementCenter(refinement_radii,\n                              dx_fine=0.001,\n                              cfl_fine=0.5,\n                              center_num=1,\n                              position=(10,0,0))\n\n# Same but with different center_num and position\ncenter2 = pg.RefinementCenter(refinement_radii,\n                              dx_fine=0.001,\n                              cfl_fine=0.5,\n                              center_num=2,\n                              position=(-10,0,0))\n\ngrid_not_synced = pg.Grid((center1, center2), outer_boundary=1000)\ngrid_synced = pg.set_dt_max_grid(grid_not_synced, dt_max=1)\n\nprint(grid_synced)\n```\nThis will output something along the lines of\n\n``` sh\nCartGrid3D::type = "coordbase"\nCarpet::domain_from_coordbase = "yes"\nCoordBase::domainsize = "minmax"\n\nDriver::ghost_size = 3\nCoordBase::boundary_size_x_lower = 3\nCoordBase::boundary_size_y_lower = 3\nCoordBase::boundary_size_z_lower = 3\nCoordBase::boundary_size_x_upper = 3\nCoordBase::boundary_size_y_upper = 3\nCoordBase::boundary_size_z_upper = 3\n\nCoordBase::xmin = 1000\nCoordBase::ymin = 1000\nCoordBase::zmin = 1000\nCoordBase::xmax = 1000\nCoordBase::ymax = 1000\nCoordBase::zmax = 1000\nCoordBase::dx = 0.64\nCoordBase::dy = 0.64\nCoordBase::dz = 0.64\nCarpet::max_refinement_levels = 8\nCarpet::time_refinement_factors = "[1,1,2,4,8,16,32,64]"\nCarpetRegrid2::num_levels_1 = 8\nCarpetRegrid2::position_x_1 = 10\nCarpetRegrid2::position_y_1 = 0\nCarpetRegrid2::position_z_1 = 0\nCarpetRegrid2::radius_1[1] = 64\nCarpetRegrid2::radius_1[2] = 32\nCarpetRegrid2::radius_1[3] = 16\nCarpetRegrid2::radius_1[4] = 8\nCarpetRegrid2::radius_1[5] = 4\nCarpetRegrid2::radius_1[6] = 2\nCarpetRegrid2::radius_1[7] = 1\nCarpetRegrid2::num_levels_2 = 8\nCarpetRegrid2::position_x_2 = -10\nCarpetRegrid2::position_y_2 = 0\nCarpetRegrid2::position_z_2 = 0\nCarpetRegrid2::radius_2[1] = 64\nCarpetRegrid2::radius_2[2] = 32\nCarpetRegrid2::radius_2[3] = 16\nCarpetRegrid2::radius_2[4] = 8\nCarpetRegrid2::radius_2[5] = 4\nCarpetRegrid2::radius_2[6] = 2\nCarpetRegrid2::radius_2[7] = 1\n```\n\nYou can also add a small shift to the grid so that the origin is not on (0,0,0)\npassing the `tiny_shift` argument to `Grid`.\n\nAnother useful method is `rl_synced_every`, which returns the number of\niterations at which all the refinement levels are at the same time.\n\n\n### Working with binary black holes\n\nProblem: you want to simulate a binary black quasi-circular merger. This can be\ntricky because you have to provide the correct linear momenta. With `Juhki`,\nthis is trivial:\n\n``` python\n#!/usr/bin/env python3\n\nfrom jhuki.twopunctures import prepare_quasicircular_inspiral\n\nmass_ratio = 1\ncoordinate_distance = 12\ndimensionless_spin_plus = (0.1, 0.2, 0.3)\ndimensionless_spin_minus = (0.4, -0.1, -0.2)\n\ntwopunctures = prepare_quasicircular_inspiral(mass_ratio,\n                                              coordinate_distance,\n                                              chi_plus=dimensionless_spin_plus,\n                                              chi_minus=dimensionless_spin_minus)\nprint(twopunctures)\n```\n\nThis will produce an output similar to the following:\n\n```\nADMBase::initial_data = twopunctures\nADMBase::initial_lapse = twopunctures-averaged\nADMBase::initial_shift = zero\nADMBase::initial_dtlapse = zero\nADMBase::initial_dtshift = zero\nTwoPunctures::give_bare_mass = no\nTwoPunctures::par_b = 6.0\nTwoPunctures::target_m_plus = 0.5\nTwoPunctures::target_m_minus = 0.5\nTwoPunctures::par_P_plus[0] = -0.000531433937072403\nTwoPunctures::par_P_plus[1] = 0.0848055056299618\nTwoPunctures::par_P_plus[2] = 0\nTwoPunctures::par_P_minus[0] = 0.000531433937072403\nTwoPunctures::par_P_minus[1] = -0.0848055056299618\nTwoPunctures::par_P_minus[2] = 0\nTwoPunctures::par_S_plus[0] = 0.025\nTwoPunctures::par_S_plus[1] = 0.05\nTwoPunctures::par_S_plus[2] = 0.075\nTwoPunctures::par_S_minus[0] = 0.1\nTwoPunctures::par_S_minus[1] = -0.025\nTwoPunctures::par_S_minus[2] = -0.05\nTwoPunctures::center_offset[0] = 0.0\nTwoPunctures::center_offset[1] = 0.0\nTwoPunctures::center_offset[2] = 0.0\n```\n\nThis module should be used along with the `Grid` one.\n\n\n## Installation\n\nThe best way to install `Jhuki` is by cloning this repo and using\n[poetry](https://python-poetry.org/). If you have poetry install, just run\n`poetry install` in the folder where you cloned the repo to install `Jhuki`.\nAlternatively, `Jhuki` is available on PyPI.\n\n## Tests\n\n`Jhuki` comes with a suite of unit tests. To run the tests,\n```sh\npoetry run pytest --cov=./ --cov-report=term\n```\nTests are automatically run after each commit by GitHub Actions. This will also\ntell you what is the test coverage.\n\n# Code style\n\n- We lint the code with `black -l 79`.\n- We work only with immutable structures.\n\n# What does _jhuki_ mean?\n\nThe word _jhuki_ belongs to the Tohono O\'odham vocabulary and means *rain*. If\n[kuibit](https://githum.com/Sbozzolo/kuibit) is the tool you use to collect the\nfruit of your `Cactus` simulations, then `jhuki` is what allowed that fruit to\nbe there in the first place.\n\n## Credits\n\nThe logo contains elements designed by [pngtree.com](pngtree.com).\n\nThe computation of the momenta for quasi-circular mergers of binary black holes\nuses\n[NRPyPN](https://einsteintoolkit.org/thornguide/EinsteinInitialData/NRPyPN/documentation.html).\nIf you use this module, please follow the citation guidelines as specified by\nthe documentation in the `NRPyPN` repo.\n',
    'author': 'Gabriele Bozzola',
    'author_email': 'gabrielebozzola@arizona.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sbozzolo/Jhuki',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
